using System;
using System.Collections.Generic;

namespace PAT.Common.Classes.Expressions.ExpressionClass
{
    public sealed class PrimitiveApplication : Expression
    {
        public const string AND = "&&";
        public const string OR = "||";
        public const string NOT = "!";

        public const string PLUS = "+";
        public const string MINUS = "-";
        public const string TIMES = "*";
        public const string DIVIDE = "/";
        public const string MOD = "mod";
        public const string NEGATIVE = "~";
        public const string NON_NEGATIVE_MOD = "non_negative_mode";

        public const string EQUAL = "==";
        public const string NOT_EQUAL = "!=";
        public const string GREATER = ">";
        public const string GREATER_EQUAL = ">=";
        public const string LESS = "<";
        public const string LESS_EQUAL = "<=";
        public const string ARRAY = ".";
        public const string ARRAYPRIME = ".'";
        public const string MIN = "Min";
        public const string MAX = "Max";

        public string Operator;
        public Expression Argument1;
        public Expression Argument2;

        public PrimitiveApplication(String op, Expression a1)
        {
            ExpressionType = ExpressionType.PrimitiveApplication;
            Operator = op;
            Argument1 = a1;
            this.HasVar = a1.HasVar;
            expressionID = Operator + Argument1.ExpressionID;             
        }

        public PrimitiveApplication(String op, Expression a1, Expression a2)
        {
            ExpressionType = ExpressionType.PrimitiveApplication;
            Operator = op;
            Argument1 = a1;
            Argument2 = a2;
            this.HasVar = a1.HasVar || a2.HasVar;

            if (Operator == MOD)
            {
                expressionID = Argument1.ExpressionID + "%" + Argument2.ExpressionID;
            }
            else
            {
                expressionID = Argument1.ExpressionID + Operator + Argument2.ExpressionID;
            }
        }

        public override String ToString()
        {
            if (Argument2 == null)
            {
                if (Operator == NEGATIVE)
                {
                    return "-(" + Argument1 + ")";
                }
                else
                {
                    return Operator + "(" + Argument1 + ")";
                }
            }
            else
            {
                if (Operator == ARRAY)
                {
                    return Argument1 + "[" + Argument2 + "]";                   
                }
                else if (Operator == MOD)
                {
                    return "(" + Argument1 + " % " + Argument2 + ")";
                }
                else
                {
                    return "(" + Argument1 + " " + Operator + " " + Argument2 + ")";
                }
            }
        }

        //public override String GetID()
        //{
        //    if (Argument2 == null)
        //    {
        //        return Operator + Argument1.GetID();                    
        //    }
        //    else
        //    {
        //        if (Operator == MOD)
        //        {
        //            return Argument1.GetID() + "%" + Argument2.GetID();
        //        }
        //        else 
        //        {
        //            return Argument1.GetID() + Operator + Argument2.GetID();
        //        }
        //    }
        //}

        
        public override List<string> GetVars()
        {           
            if (Operator == ARRAY)
            {
                List<string> vars = new List<string>(16);
                vars.Add(Argument1.ToString());
                if (Argument2 != null)
                {
                    Ultility.Ultility.AddList(vars, this.Argument2.GetVars());
                }
                return vars;
            }
            else
            {
                List<string> vars = this.Argument1.GetVars();
                if (Argument2 != null)
                {
                    Ultility.Ultility.AddList(vars, this.Argument2.GetVars());
                }
                return vars;
            }            
        }

        public override Expression ClearConstant(Dictionary<string, Expression> constMapping)
        {
            if (Argument2 != null)
            {
                return new PrimitiveApplication(Operator, Argument1.ClearConstant(constMapping), Argument2.ClearConstant(constMapping));
            }
            else
            {
                return new PrimitiveApplication(Operator, Argument1.ClearConstant(constMapping));
            }
        }



        public bool IsBinaryExpression()
        {
            return (this.Argument2 != null);
        }

        public bool IsBinaryLogicalExpression()
        {
            return (this.Operator == AND || this.Operator == OR);
        }

        public bool IsBinaryNumericalExpression()
        {
            return (this.Operator == PLUS || this.Operator == MINUS || this.Operator == TIMES ||
                        this.Operator == DIVIDE || this.Operator == MOD || this.Operator == NON_NEGATIVE_MOD || this.Operator == MIN || this.Operator == MAX);
        }

        public bool IsRelationalExpression()
        {
            return (this.Operator == EQUAL || this.Operator == NOT_EQUAL || this.Operator == GREATER || this.Operator == GREATER_EQUAL ||
                        this.Operator == LESS || this.Operator == LESS_EQUAL);
        }

        public bool IsArrayExpression()
        {
            return (this.Operator == ARRAY) || (this.Operator == ARRAYPRIME);
        }

        public bool IsBooleanExpression()
        {
            return IsBinaryLogicalExpression() || IsRelationalExpression();
        }


        public override bool HasExternalLibraryCall()
        {
            if(IsBinaryExpression())
            {
                return Argument1.HasExternalLibraryCall() || Argument2.HasExternalLibraryCall();
            }
            else
            {
                return Argument1.HasExternalLibraryCall();
            }
        }
    }
}