﻿using System.Collections.Generic;
using PAT.Common;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.SemanticModels.LTS.Assertion;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;

namespace PAT.PAR.Assertions
{
    public class PARAssertion
    {
        public static void Initialize(AssertionBase Assertion, DefinitionRef Process, SpecificationBase spec)
        {
            Specification Spec = spec as Specification;
            //get the relevant global variables; remove irrelevant variables so as to save memory;
            Valuation GlobalEnv = Spec.SpecValuation.GetVariableChannelClone(Process.GetGlobalVariables(), Process.GetChannels());
            //Initialize InitialStep
            Assertion.InitialStep = new PARConfiguration(Process, Constants.INITIAL_EVENT, null, GlobalEnv, false, false, Constraint.constraintWithNonNegativeParameters()); // new DBM() replaced by constraint // trueConstraint()
            (Assertion.InitialStep as PARConfiguration).SetTimer();
            Assertion.MustAbstract = Process.MustBeAbstracted();
        }

        public static void Initialize(AssertionRefinement Assertion, DefinitionRef ImplementationProcess, DefinitionRef SpecificationProcess, SpecificationBase spec, byte InitTimerID)
        {
            Specification Spec = spec as Specification;

            //get the relevant global variables; remove irrelevant variables so as to save memory;
            List<string> varList = ImplementationProcess.GetGlobalVariables();
            Valuation GlobalEnv = Spec.SpecValuation.GetVariableChannelClone(varList, ImplementationProcess.GetChannels());

            Assertion.InitialStep = new PARConfiguration(ImplementationProcess, Constants.INITIAL_EVENT, null, GlobalEnv, false, false, Constraint.constraintWithNonNegativeParameters()); // new DBM() replaced by constraint // trueConstraint
            (Assertion.InitialStep as PARConfiguration).SetTimer();

            //get the relevant global variables; remove irrelevant variables so as to save memory;
            varList = SpecificationProcess.GetGlobalVariables();
            GlobalEnv = Spec.SpecValuation.GetVariableChannelClone(varList, SpecificationProcess.GetChannels());

            Assertion.InitSpecStep = new PARConfiguration(SpecificationProcess, Constants.INITIAL_EVENT, null, GlobalEnv, false, false, Constraint.constraintWithNonNegativeParameters()); // new DBM() replaced by constraint // trueConstraint
            (Assertion.InitSpecStep as PARConfiguration).SetTimer();
            Assertion.MustAbstract = ImplementationProcess.MustBeAbstracted();

            if (SpecificationProcess.MustBeAbstracted())
            {
                throw new ParsingException(
                    "Process " + Assertion.SpecProcess +
                    " has infinite states and therefore can not be used as a property model for refinement checking!", Assertion.AssertToken);
            }
        }
    }
}
