using System.Collections.Generic;
using PAT.Common;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.SemanticModels.LTS.Assertion;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;
using PAT.PAR.Ultility;

namespace PAT.PAR.Assertions
{
    public class PARAssertionReachabilityWith : AssertionReachabilityWith
    {
        private DefinitionRef Process;

        public PARAssertionReachabilityWith(DefinitionRef processDef, string reachableState, QueryConstraintType cont, Expression constraintCondition)
            : base(reachableState, cont, constraintCondition)
        {
            Process = processDef;
        }

        // ETIENNE SAYS: WHY IS THIS METHOD MUCH LONGER THAN IN RTS?
        public override void Initialize(SpecificationBase spec)
        {
            base.Initialize(spec);

            Specification Spec = spec as Specification;
            ReachableStateCondition = Spec.DeclarationDatabase[ReachableStateLabel];

            List<string> varList = Process.GetGlobalVariables();
            varList.AddRange(ReachableStateCondition.GetVars());
            varList.AddRange(ConstraintCondition.GetVars());                        

            Valuation GlobalEnv = Spec.SpecValuation.GetVariableChannelClone(varList, Process.GetChannels());
            
            //Debug.Assert(StartingProcess is DefinitionRef);

            //Initialize InitialStep
            InitialStep = new PARConfiguration(Process, Constants.INITIAL_EVENT, null, GlobalEnv, false, false, Constraint.constraintWithNonNegativeParameters()); // trueConstraint

            (InitialStep as PARConfiguration).SetTimer();

            MustAbstract = Process.MustBeAbstracted();

            if (MustAbstract)
            {
                throw new ParsingException(
                    "Process " + StartingProcess +
                    " has infinite states and therefore can not be used to assert reachability with MIN/MAX constraints!",
                    AssertToken);

            }

        }

        //public override void RunVerification()
        //{
        //    if (!Process.IsTime())
        //    {
        //        base.RunVerification();
        //    }
        //    else
        //    {
        //        TimedVerification();
        //    }
        //}

        //private void TimedVerification()
        //{
        //    StringDictionary<List<Constraint>> Visited = new StringDictionary<List<Constraint>>(Common.Classes.Ultility.Ultility.MC_INITIAL_SIZE);

        //    Expression conditionExpression = this.ReachableStateCondition;

        //    Stack<PARConfiguration> working = new Stack<PARConfiguration>(1024);

        //    //DataStore.DataManager.Visit(InitialStep.GetID());
        //    PARUltility.Visit(Visited, (InitialStep as PARConfiguration));

        //    working.Push(InitialStep as PARConfiguration);
        //    Stack<int> depthStack = new Stack<int>(1024);
        //    depthStack.Push(0);

        //    //List<string> procPath = new List<string>(1024);
        //    List<int> depthList = new List<int>(1024);

        //    ExtremValue = null;
        //    ReachableCount = 0;
        //    List<ConfigurationBase> CounterExampleTraceLocal = new List<ConfigurationBase>();

        //    do
        //    {
        //        if (CancelRequested)
        //        {
        //            VerificationOutput.NoOfStates = Visited.Count + Visited.ExtraCounter;
        //            VerificationOutput.NoOfGroupedStates = Visited.ExtraCounter;
        //            return;
        //        }

        //        PARConfiguration current = working.Pop();
        //        int depth = depthStack.Pop();

        //        while (depth > 0 && depthList[depthList.Count - 1] >= depth)
        //        {
        //            int lastIndex = depthList.Count - 1;
        //            depthList.RemoveAt(lastIndex);
        //            //procPath.RemoveAt(lastIndex);
        //            CounterExampleTraceLocal.RemoveAt(lastIndex);
        //        }

        //        CounterExampleTraceLocal.Add(current);

        //        ExpressionValue v = EvaluatorDenotational.Evaluate(conditionExpression, current.GlobalEnv);
        //        if ((v as BoolConstant).Value)
        //        {
        //            //VerificationOutput.VerificationResult = VerificationOutput.VerificationResultType.VALID;
        //            //VerificationOutput.NoOfStates = Visited.Count + Visited.ExtraCounter;
        //            //NoOfGroupedStates = Visited.ExtraCounter;
        //            //return;
        //            ReachableCount++;
        //            v = EvaluatorDenotational.Evaluate(ConstraintCondition, current.GlobalEnv);
        //            if (v is IntConstant)
        //            {
        //                int value = (v as IntConstant).Value;
        //                if (ExtremValue == null)
        //                {
        //                    ExtremValue = value;
        //                    VerificationOutput.CounterExampleTrace = new List<ConfigurationBase>(CounterExampleTraceLocal);
        //                }
        //                else
        //                {
        //                    switch (Contraint)
        //                    {
        //                        case QueryConstraintType.MAX:
        //                            if (value > ExtremValue.Value)
        //                            {
        //                                ExtremValue = value;
        //                                VerificationOutput.CounterExampleTrace = new List<ConfigurationBase>(CounterExampleTraceLocal);
        //                            }
        //                            break;
        //                        case QueryConstraintType.MIN:
        //                            if (value < ExtremValue.Value)
        //                            {
        //                                ExtremValue = value;
        //                                VerificationOutput.CounterExampleTrace = new List<ConfigurationBase>(CounterExampleTraceLocal);
        //                            }
        //                            break;
        //                    }
        //                }
        //            }
        //        }

        //        depthList.Add(depth);
        //        List<PARConfiguration> list = current.MakeOneMoveLocal();
        //        VerificationOutput.Transitions += list.Count;

        //        for (int i = list.Count - 1; i >= 0; i--)
        //        {
        //            PARConfiguration step = list[i];
        //            //if (!DataStore.DataManager.Visit(step.GetID()))

        //            if (!PARUltility.Visit(Visited, step))
        //            {
        //                working.Push(step);
        //                depthStack.Push(depth + 1);
        //            }
        //        }
        //    } while (working.Count > 0);

        //    if (ExtremValue == null)
        //    {
        //        VerificationOutput.VerificationResult = VerificationResultType.INVALID;
        //    }
        //    else
        //    {
        //        VerificationOutput.VerificationResult = VerificationResultType.VALID;
        //    }

        //    VerificationOutput.NoOfStates = Visited.Count + Visited.ExtraCounter;
        //    VerificationOutput.NoOfGroupedStates = Visited.ExtraCounter;
        //    return;

        //}

        public override string StartingProcess
        {
            get
            {
                return Process.ToString();
            }
        }
    }
}