//#define DEBUG_PRINT

using System.Collections;
// For dictionnary and queue
using System.Collections.Generic;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.SemanticModels.LTS.Assertion;
// For StringBuilder
using System.Text;
using PAT.Common.Classes.Ultility;
using PAT.PAR.LTS;

namespace PAT.PAR.Assertions
{
    public class PARSynthesisReachabilityAll : AssertionReachability
    {
        private DefinitionRef Process;
        // Number of (finally) reachable states
        private int nbStates;
        // Number of states visited (but potentially merged)
        private int nbVisited;
        // Number of transitions (not represented explicitely so far)
        private int nbTransitions;
        // Maximum number of clocks met during the analysis
        private int maxNbClocks;

        private bool INCLUSION_TEST = false;


        public PARSynthesisReachabilityAll(DefinitionRef processDef)
            : base("")
        {
            Process = processDef;
        }

        public override void Initialize(SpecificationBase spec)
        {
//            base.Initialize(spec);

  //          Specification Spec = spec as Specification;

            ModelCheckingOptions = new ModelCheckingOptions();
            List<string> DeadlockEngine = new List<string>();
            DeadlockEngine.Add(Constants.ENGINE_WITH_PAR_OPTIM);
            DeadlockEngine.Add(Constants.ENGINE_WITHOUT_PAR_OPTIM);
            ModelCheckingOptions.AddAddimissibleBehavior(Constants.COMPLETE_BEHAVIOR, DeadlockEngine);


            //ReachableStateCondition = Spec.DeclarationDatabase[ReachableStateLabel];
            PARAssertion.Initialize(this, Process, spec);

        }

        // ETIENNE SAYS: What's that?!
        public override string StartingProcess
        {
            get
            {
                return Process.ToString();
            }
        }



        public override string ToString()
        {
            return StartingProcess + " reachesall";
        }

        /// <summary>
        /// Run the verification and get the result.
        /// </summary>
        /// <returns></returns>

        /************************************************************
        * MAIN IMPLEMENTATION OF THE FULL REACHABILITY ALGORITHM
        ************************************************************/
         public override void RunVerification()
         {
             // Cast for simplicity purpose
             PARConfiguration InitialState = InitialStep as PARConfiguration;

             // Define the hash for states already visited
             Dictionary<string, PARConfiguration> Visited = new Dictionary<string, PARConfiguration>();

             // Define the hash discrete_id => list of config_id
             // (associate to each discrete part of a configuration a list of configurations id having this discrete part)
             // (useful for retrieving states with a given process and values for variables)
             Dictionary<string, List<int>> SimilarConfigurations = new Dictionary<string, List<int>>();

             // Define the hash int_id => config_id
             Dictionary<int, string> VisitedIds = new Dictionary<int, string>();

             // Define the queue of states to be visited
             // TO DO: replace queue by set, to avoid unnecessary duplication of states
             Queue<PARConfiguration> queue = new Queue<PARConfiguration>();
             queue.Enqueue(InitialState);

             // TO DO: keep track of the transitions, for presentation purpose
             nbTransitions = 0;

             // No state visited yet
             nbVisited = 0;

             // No clock yet
             maxNbClocks = 0;

             // For giving an int ID to configurations
             int CurrentId = 0;

             // Define the options
             if (SelectedEngineName == Constants.ENGINE_WITH_PAR_OPTIM)
             {
                 Constraint.EQUIVALENCE_REDUCTION = true;
             }
             else
             {
                 Constraint.EQUIVALENCE_REDUCTION = false;
             }



 #if DEBUG_PRINT
             System.Console.Out.WriteLine("************************************************************");
             System.Console.Out.WriteLine("STARTING FULL REACHABILITY COMPUTATION");
             System.Console.Out.WriteLine("Initial configuration: \n" + InitialState.ToString());
#endif

             while (queue.Count > 0)
             {
                 // Consider one more state
                 PARConfiguration current_state = queue.Dequeue();

                 // Count to be visited
                 nbVisited++;

                 //System.Console.Out.WriteLine("***********************************************");
                 //System.Console.Out.WriteLine("NB STATES = " + Visited.Count);
                 //System.Console.Out.WriteLine("NB VISITED = " + nbVisited);
                 //System.Console.Out.WriteLine("***********************************************");


 #if DEBUG_PRINT
                 System.Console.Out.WriteLine("************************************************************");
                 System.Console.Out.WriteLine("Dequeued following state:");
                 System.Console.Out.WriteLine(current_state);
 #endif

                 // Compute the id
                 string id = current_state.GetID();

                 // Test whether the state was visited before
                 bool visited = Visited.ContainsKey(id);

                 // Otherwise check if we have state inclusion
                 if (!visited && INCLUSION_TEST)
                 {
                     // Retrieve the discrete id
                     string discrete_id = current_state.GetDiscreteID();
                     // Retrieve the constraint
                     Constraint current_constraint = current_state.TheConstraint;

                     // Find all int_ids corresponding to this discrete id
                     List<int> int_ids = SimilarConfigurations[discrete_id];
                     // Iterate on all similar configurations
                     foreach (int int_id in int_ids)
                     {
                         // Retrieve the string id
                         string string_id = VisitedIds[int_id];
                         // Retrieve the real configuration
                         PARConfiguration config = Visited[string_id];
                         Constraint constraint_for_comparison = config.TheConstraint;

                         // Test inclusion
                         // TODO !
                         // if(constraint_for_comparison.isIncluded(......

                     }
                     // TO DO
                     visited = false;
                 }

                 // TODO: update transition graph!
                 nbTransitions++;


                     // TO DO: case option inclusion!
                 if (visited)
                 {
 #if DEBUG_PRINT
                     System.Console.Out.WriteLine("State already visited. Skip.");
 #endif
                 }
                 // If new: add it
                 else
                 {
                     // Add to the list of visited states
                     Visited.Add(id, current_state);


                     if (INCLUSION_TEST)
                     {
                         // Assign an int id to this new configuration
                         int int_id = CurrentId;
                         // Increment for the next id
                         CurrentId++;


                         // Update VisitedIds
                         VisitedIds[int_id] = id;


                         // Update SimilarConfigurations
                         // 1) Retrieve discrete id, i.e., without the constraint
                         string discrete_id = current_state.GetDiscreteID();
                         // 2) Check whether there is already a mapping discrete_id in SimilarConfigurations
                         if (!SimilarConfigurations.ContainsKey(discrete_id))
                         {
                             // Create the empty list
                             SimilarConfigurations[discrete_id] = new List<int>();
                         }
                         // 3) Add the current id
                         SimilarConfigurations[discrete_id].Add(int_id);
                     }


                     // Update the maximum number of clocks
                     int current_nb_clocks = current_state.getNbClocks();
                     if (current_nb_clocks > maxNbClocks)
                     {
                         maxNbClocks = current_nb_clocks;
                     }

                     // Compute successors
                     PARConfiguration[] states = current_state.MakeOneMove() as PARConfiguration[];
                     // Add them to the queue
                     foreach (PARConfiguration state in states)
                     {
                         queue.Enqueue(state);
 #if DEBUG_PRINT
                                 System.Console.Out.WriteLine("\n--New state added to queue (now " + queue.Count + " states in queue): \n" + state);
 #endif
                     }
                 }

             } // while queue not empty

             // END OF THE ALGORITHM
             // Store
             nbStates = Visited.Count;

             // Update result
             this.VerificationOutput.NoOfStates = Visited.Count;
             this.VerificationOutput.Transitions = nbTransitions;
             // No result here
             // this.VerificationOutput.VerificationResult = VerificationResultType.VALID;


             if (INCLUSION_TEST)
             {
                 System.Console.Out.WriteLine("------------------------------------------------------------");
                 System.Console.Out.WriteLine(" SIMILAR CONFIGURATIONS");
                 foreach (KeyValuePair<string, List<int>> kvp in SimilarConfigurations)
                 {
                     // Retrieve the discrete id
                     string discrete_id = kvp.Key;
                     // Retrieve the list of full ids
                     List<int> full_ids = kvp.Value;
                     System.Console.Out.WriteLine("------------------------------------------------------------");
                     System.Console.Out.WriteLine(discrete_id);
                     foreach (int int_id in full_ids)
                     {
                         System.Console.Out.WriteLine("  - " + int_id + ": " + VisitedIds[int_id]);
                     }
                 }
             }



#if DEBUG_PRINT
             System.Console.Out.WriteLine("************************************************************");
             System.Console.Out.WriteLine("REACHABILITY ANALYSIS SUCCESSFULLY TERMINATED");
             System.Console.Out.WriteLine(Visited.Count + " states reached: ");
             foreach (KeyValuePair<string, PARConfiguration> kvp in Visited)
             {
                 System.Console.Out.WriteLine("------------------------------------------------------------");
                 System.Console.Out.WriteLine("Id=" + kvp.Key);
                 System.Console.Out.WriteLine(kvp.Value);
             }
             System.Console.Out.WriteLine("************************************************************");
 #endif

         }

         public override string GetResultString()
         {
             StringBuilder sb = new StringBuilder();
             sb.AppendLine("********Full reachability result********");

             if (SelectedEngineName == Constants.ENGINE_WITH_PAR_OPTIM)
             {
                 sb.AppendLine("Algorithm version: with state equivalence reduction [ALSD12].");
             }
             else
             {
                 sb.AppendLine("Algorithm version: without state equivalence reduction [ALSD12].");
             }
             
             sb.AppendLine(nbStates + " states reachable.");
//             sb.AppendLine("These states were visited " + nbVisited + " times.");
//             sb.AppendLine("Number of transitions: " + nbTransitions);
             sb.AppendLine("Maximal number of clocks: " + maxNbClocks);
             return sb.ToString();
         }
    }
}