using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class Assertion : PARProcess
    {
        public Expression ConditionalExpression;
        public int LineNumber;

        public Assertion(Expression conditionExpression, int line)
        {
            ConditionalExpression = conditionExpression;
            LineNumber = line;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.ASSERTION + conditionExpression.ExpressionID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            ExpressionValue v = EvaluatorDenotational.Evaluate(ConditionalExpression, eStep.GlobalEnv);

            if ((v as BoolConstant).Value)
            {
                returnList.Add(new PARConfiguration(new Stop(), Constants.TERMINATION, null, eStep.GlobalEnv, false, false, eStep.TheConstraint));
            }
            else
            {
                throw new RuntimeException("Assertion at line " + LineNumber + " failed: " + ConditionalExpression.ToString());
            }

            return returnList;
        }

        public override string ToString()
        {
            return " assert(" + ConditionalExpression + ")";
        }

        public override List<string> GetGlobalVariables()
        {
            return ConditionalExpression.GetVars();
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression newCon = ConditionalExpression.ClearConstant(constMapping);

            return new Assertion(newCon, LineNumber);
        }
    }
}