using System;
using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class ChannelOutput : PARProcess
    {
        public string ChannelName;
        public Expression[] ExpressionList;
        public PARProcess Process;
        public bool IsUrgent;

        public ChannelOutput(string evtName, Expression[] e, PARProcess process, bool isUrgent)
        {
            ChannelName = evtName;
            Process = process;
            ExpressionList = e;
            IsUrgent = isUrgent;
            String ID = ChannelName + "!" + Common.Classes.Ultility.Ultility.PPIDListDot(ExpressionList) + (IsUrgent ? Constants.EVENTPREFIX_URGENT : Constants.EVENTPREFIX) + Process.ProcessID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> list = new List<PARConfiguration>();
            Valuation globalEnv = eStep.GlobalEnv;
            ChannelQueue Buffer = null;

            if (globalEnv.Channels.TryGetValue(this.ChannelName, out Buffer))
            {
                if (Buffer.Count < Buffer.Size)
                {
                    ExpressionValue[] values = new ExpressionValue[ExpressionList.Length];
                    string eventName = ChannelName + "!";
                    string eventID = ChannelName + "?";

                    for (int i = 0; i < ExpressionList.Length; i++)
                    {
                        values[i] = EvaluatorDenotational.Evaluate(ExpressionList[i], globalEnv);
                        if (i == ExpressionList.Length - 1)
                        {
                            eventName += values[i].ToString();
                            eventID += values[i].ExpressionID;
                        }
                        else
                        {
                            eventName += values[i].ToString() + ".";
                            eventID += values[i].ExpressionID + ".";
                        }
                    }

                    ChannelQueue newBuffer = Buffer.Clone();
                    newBuffer.Enqueue(values);
                    globalEnv = globalEnv.GetChannelClone();
                    globalEnv.Channels[ChannelName] = newBuffer;

                    list.Add(new PARConfiguration(Process, eventID, eventID != eventName ? eventName : null, globalEnv,
                                                  false, IsUrgent, eStep.TheConstraint));
                }
            }

            return list;
        }



        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> list = new List<PARConfigurationWithChannelData>(1);
            if (SpecificationBase.SyncrhonousChannelNames.Contains(ChannelName))
            {
                string eventName = ChannelName;
                string eventID = ChannelName;

                Expression[] newExpressionList = new Expression[ExpressionList.Length];

                for (int i = 0; i < ExpressionList.Length; i++)
                {
                    newExpressionList[i] = EvaluatorDenotational.Evaluate(ExpressionList[i], eStep.GlobalEnv);
                    eventName += "." + newExpressionList[i];
                    eventID += "." + newExpressionList[i].ExpressionID;
                }

                list.Add(new PARConfigurationWithChannelData(Process, eventName, eventID, eStep.GlobalEnv, false, IsUrgent, ChannelName, newExpressionList, eStep.TheConstraint, eStep.TheConstraint));
            }

            return list;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {         
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> toReturn = Process.GetGlobalVariables();

            foreach (Expression expression in ExpressionList)
            {
                Common.Classes.Ultility.Ultility.AddList(toReturn, expression.GetVars());
            }

            return toReturn;
        }

        public override List<string> GetChannels()
        {
            List<string> toReturn = Process.GetChannels();

            if (!toReturn.Contains(ChannelName))
            {
                toReturn.Add(ChannelName);
            }

            return toReturn;
        }

        public override string ToString()
        {
            return ChannelName + "!" + Common.Classes.Ultility.Ultility.PPStringListDot(ExpressionList).TrimStart('.') + (IsUrgent ? "->>" : "->") + Process;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression[] newExpression = new Expression[ExpressionList.Length];
            for (int i = 0; i < ExpressionList.Length; i++)
            {
                newExpression[i] = ExpressionList[i].ClearConstant(constMapping);
            }
            
            return new ChannelOutput(ChannelName, newExpression, Process.ClearConstant(constMapping), IsUrgent);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override bool IsTime()
        //{
        //    return Process.IsTime();
        //}

    }  
}