using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class ConditionalChoiceBlocking : PARProcess
    {
        public PARProcess FirstProcess;
        public Expression ConditionalExpression;

        public ConditionalChoiceBlocking(PARProcess firstProcess, Expression conditionExpression)
        {
            FirstProcess = firstProcess;
            ConditionalExpression = conditionExpression;

            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.CONDITIONAL_CHOICE + conditionExpression.ExpressionID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            ExpressionValue v = EvaluatorDenotational.Evaluate(ConditionalExpression, eStep.GlobalEnv);

            if ((v as BoolConstant).Value)
            {
                // REMOVED BY ETIENNE (hopefully this is handled through the urgency in PARConfiguration)
                // eStep.AddInequality(eStep.NewClock, TimerOperationType.Equals, 0);
                returnList.Add(new PARConfiguration(FirstProcess, Constants.TAU, "[ifb(" + ConditionalExpression + ")]", eStep.GlobalEnv, false, true, eStep.TheConstraint));
            }

            return returnList;
        }
        

        public override string ToString()
        {
            return "ifb " + ConditionalExpression + " {" + FirstProcess.ToString() + "}";
        }



        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return FirstProcess.GetAlphabets(visitedDefinitionRefs);
        }



        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, ConditionalExpression.GetVars());

            return Variables;
        }

        public override List<string> GetChannels()
        {
            return FirstProcess.GetChannels();
        }


        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            Expression newCon = ConditionalExpression.ClearConstant(constMapping);         
            PARProcess newFirstProc = FirstProcess.ClearConstant(constMapping);
            
            return new ConditionalChoiceBlocking(newFirstProc, newCon);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted();
        }

        //public override bool IsTime()
        //{
        //    return FirstProcess.IsTime();
        //}
    }
}