﻿//#define DEBUGCONSOLE

using System.Linq;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Text.RegularExpressions;
using PAT.Common.Classes.Ultility;
using PAT.Common.Classes.Expressions.ExpressionClass;
using System.Runtime.InteropServices;
using System.IO;
using System.Windows.Forms;



namespace PAT.PAR.LTS
{
    // The operators
    public enum TimerOperationType : byte
    {
        Equals,
        LessThanOrEqualTo,
        GreaterThanOrEqualTo,
        LessThan,
        GreaterThan
    }


    public class Constraint
    {
        /* CONSTANTS */

        // SHOULD NOT BE CHANGED; in particular, the renaming algorithm highly depend on FIRST_CLOCK = 0
        private const int FIRST_CLOCK = 1;


        /* MEMBERS */

        // The list of clocks
        private List<byte> ClockList;

        // Object coming from PPL
        // DEBUG ETIENNE : public
        public Polyhedron thePolyhedron;


        /* STATIC MEMBERS */

        // True constraint with non-negative parameters
        private static Polyhedron NonNegativeParametersConstraint = null;
        // Array id -> param
        private static string[] ParameterNames;
        // Hash table param -> id
        private static Dictionary<string, byte> ParameterIds;
        public static byte NbParameters = 0;

        // Some strings (can be changed)
        public static string STRING_CLOCK_NAME      = "x";
        public static string STRING_VAR_NAME_TMP    = "##";
        public static string STRING_SEP_SET = ",";
        // Some strings (depend on the C++ interface; do not change!)
        public static string STRING_VAR_NAME = "v";
        public static string STRING_SEP_AND = "&";
        //public static string STRING_TRUE = "true";
        //public static string STRING_FALSE           = "false";
        //public static string STRING_OP_LEQ          = "<=";
        //public static string STRING_OP_EQ           = "=";
        //public static string STRING_OP_L            = "<";

        public static bool EQUIVALENCE_REDUCTION = true;

        // Get the id of a parameter
        static private byte GetParameterId(string ParameterName)
        {
            byte parameter_id;
            ParameterIds.TryGetValue(ParameterName, out parameter_id);
            return parameter_id;
        }


        // Translate a normal clock id into a clock id for a constraint
        static private byte ClockIdForConstraint(byte ClockID)
        {
//#if DEBUGCONSOLE
//            System.Console.Out.WriteLine("          ClockIdForConstraint: adding '"
//                + NbParameters + "' to '" + ClockID + "'; result: '" + ((byte) (ClockID + NbParameters)) + "'");
//#endif
            return (byte) (ClockID + NbParameters - FIRST_CLOCK);
        }


        //// Translate a normal clock id from a clock id in a constraint
        //static private byte ClockIdFromConstraint(byte ClockID)
        //{
        //    return (byte) (ClockID - NbParameters);
        //}

 

        private Constraint()
        {
            //RunPPLTests();
            
            // Initialize the list of clocks
            ClockList = new List<byte>();

            thePolyhedron = new Polyhedron();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    (New empty constraint)" + this);
#endif
        }


         // Static method to compute (only once per model) the non-negative parameters constraint, as well as the correspondence id <-> param
        public static void InitParameters()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("^^^^^^^^^^^^^^^^^^^^ \n INITIALIZING PARAMETERS ");
#endif

            // Get the number of parameters
            NbParameters = (byte) Specification.Parameters.Count();

            // Create array id -> param
            ParameterNames = new string[NbParameters];

            // Create hash table param -> id
            ParameterIds = new Dictionary<string, byte>(NbParameters);

            // Create polyhedron
            Polyhedron p = new Polyhedron();

            // Iterate on the parameters
            byte param_index = 0;
            foreach (string name in Specification.Parameters)
            {
                // Update array id -> param
                ParameterNames[param_index] = name;

                // Update hash table param -> id
                ParameterIds.Add(name, param_index);

                // Add inequality
                p.AddGeqConst(param_index, 0);

                // Increment param id
                param_index++;
            }
            // Copy to the static field
            NonNegativeParametersConstraint = p;

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("INITIALIZED PARAMETERS: " + NonNegativeParametersConstraint.GetId() + " \n ^^^^^^^^^^^^^^^^^^^^");
#endif
            
        }

        /************************************************************
         * CONSTRUCTORS
         ***********************************************************/

       /**
         * Creates a new constraint, with a set of inequalities 'p >= 0', for all parameters 'p'
         */
        public static Constraint constraintWithNonNegativeParameters()
        {
            // Create and update the constraint
            Constraint c = new Constraint();
            c.thePolyhedron = NonNegativeParametersConstraint.Clone();
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    New constraint: \n" + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif
            return c;
        }



        ///**
        // * Create a new constraint on parameters only (no clock)
        // */
        //private static Constraint ConstraintOnParameters(string TheConstraintOnParam)
        //{
        //    Constraint result = new Constraint();
        //    // TO DO
        //    //result.thePolyhedron = NonNegative .Clone();
        //    return result;
        //}



        /**
         * Creates a new constraint under the form of an inequality clockID <= constant
         */
        public static Constraint inequality(byte clockID, int constant)
        {
            // Create the constraint
            Constraint c = new Constraint();
            // Add the inequality
            c.thePolyhedron.AddLeqConst(ClockIdForConstraint(clockID), constant);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    New inequality: " + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif
            return c;
        }



        /**
         * Creates a new constraint under the form of a parametric inequality clockID <= parameter
         */
        public static Constraint parInequality(byte clockID, string parameter)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Entering Constraint::parInequality(" + clockID + ", " + parameter + ")");
#endif
            // Create the constraint
            Constraint c = new Constraint();
            // Get parameter id
            byte parameter_id = GetParameterId(parameter);
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Found id '" + parameter_id + "' for parameter '" + parameter + "'");
#endif
            // Add the inequality
            c.thePolyhedron.AddLeqVar(ClockIdForConstraint(clockID), parameter_id);
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("      Adding LeqVar: " + ClockIdForConstraint(clockID) + " <= " + parameter_id);
            System.Console.Out.WriteLine("    New par inequality: " + c.ToString());
            System.Console.Out.WriteLine("    ====================");
#endif

            return c;
        }



        /************************************************************
        * ACCESS WITH NO MODIFICATION
        ***********************************************************/
        

        public int GetNbClocks()
        {
            return ClockList.Count();
        }



        public bool HasClocks()
        {
            return ClockList.Count() != 0;
        }


        public bool ContainsClock(byte clockID)
        {
            return ClockList.Contains(clockID);
        }



        /** 
         * Generate a new clock ID if not present yet
         */
        public byte GetNewClockID()
        {
            // Case no clock: first clock index
            if (ClockList.Count() == 0)
            {
                return FIRST_CLOCK;
            }

            byte i = FIRST_CLOCK;
            while (true)
            {
                if (ClockList.Contains(i))
                {
                    i++;
                }
                else
                {
                    return i;
                }
            }

            // Code unreachable
            throw new Exception("This code shouldn't be reached in GetNewClockID()");
        }



        public bool IsSatisfiable()
        {
            return thePolyhedron.IsSatisfiable();
        }



        public List<byte> GetNewClocks()
        {
            if (ClockList.Count() == 0)
            {
                return new List<byte>();
            }

            // No Clone!
            return ClockList;
        }


        private bool HasNoClocks()
        {
            // Either true
            return thePolyhedron.IsTrue()
                // Or False
                || thePolyhedron.IsFalse()
                // Or the greatest clock is larger than the number of parameters
                || thePolyhedron.GreatestVariable() < NbParameters
            ;
        }


        //// Check if a constraint satisfies an inequality "clock ~ constant"
        //public bool ImpliesInequality(byte clockID, TimerOperationType op, int constant)
        //{
        //    // TO DO
        //    //string inequality = ClockToString(clockID) + opToString(op) + constant.ToString();
        //    return false; // PPL.contains(inequality, this.theConstraint);
        //}




        //// Check if a constraint satisfies a parametric inequality "clock ~ parameter"
        //public bool ImpliesParInequality(byte clockID, TimerOperationType op, string parameter)
        //{
        //    // TO DO
        //    //string inequality = ClockToString(clockID) + opToString(op) + parameter;
        //    return false; //  PPL.contains(inequality, this.theConstraint);
        //}


        // Retrieve the value associated to a parameter; no test performed!
        private static int EvaluateParameter(string parameter, Dictionary<string, int> ParameterValuation)
        {
            int value = 0;
            ParameterValuation.TryGetValue(parameter, out value);
            return value;
        }


//        // Parse an inequality, evalute it w.r.t. the reference valuation; if incompatible, negate it and return it
//        // If compatible, return null
//        private static Constraint NegateInequalityIfIncompatible(string inequality, Dictionary<string, int> ParameterValuation)
//        {
//#if DEBUGCONSOLE
//            System.Console.Out.WriteLine("      Entering NegateInequalityIfIncompatible(" + inequality + ")");
//#endif
//            // Get the operator of the inequality
//            TimerOperationType op;
//            if (inequality.Contains(STRING_OP_LEQ))
//            {
//                op = TimerOperationType.LessThanOrEqualTo;
//            }
//            else if (inequality.Contains(STRING_OP_L))
//            {
//                op = TimerOperationType.LessThan;
//            }
//            else if (inequality.Contains(STRING_OP_EQ))
//            {
//                op = TimerOperationType.Equals;
//            }
//            else
//            {
//                throw new Exception("Inequality " + inequality + " is not well-formed: no operator found.");
//            }

//            // Remove '<=' and split
//            string[] split = inequality.Split(new Char[] { '<', '=' });

//            // Retrieve the members
//            string member_left = "";
//            string member_right = "";
//            // Depend on the form of the split
//            if (split.Length == 2){
//                member_left = split[0];
//                member_right = split[1];
//            }
//            else if (split.Length == 3){
//                member_left = split[0];
//                member_right = split[2];
//            }
//            // Check that the form is correct
//            if (member_left == "" || member_right == "")
//            {
//                string error = "Inequality " + inequality + " is not well-formed.\n Split = ";
//                for(int i = 0; i < split.Length; i++)
//                {
//                    error += i + " : " + split[i] + " - ";
//                }
//                throw new Exception(error);
//            }

//            List<string> members = new List<string>();
//            members.Add(member_left);
//            members.Add(member_right);

//            // To differentiate between left and right member
//            bool left = true;
//            // To compute left and right evaluation
//            int left_valuation = 0;
//            int right_valuation = 0;
//            foreach (string member in members)
//            {
//#if DEBUGCONSOLE             
//                System.Console.Out.WriteLine("      Split member " + member);
//#endif
//                // Split w.r.t. '+'
//                string[] terms = member.Split(new Char[] { '+' });

//                foreach (string term in terms)
//                {
//                    if (term != "")
//                    {
//#if DEBUGCONSOLE
//                System.Console.Out.WriteLine("        Split term " + term);
//#endif

//                        // Look for the initial digits
//                        Regex pattern = new Regex(@"\b(\d*)");
//                        Match m = pattern.Match(term);
//                        if (m.Success && m.Length != 0)
//                        {
//                            string digits = m.Groups[1].ToString();
//                            // If only digit: constant
//                            if (digits.Length == term.Length)
//                            {
//                                // Update term
//                                if (left)
//                                {
//                                    left_valuation += Convert.ToInt32(digits);
//#if DEBUGCONSOLE
//                                    System.Console.Out.WriteLine("        Adding constant on the left: " + digits);
//#endif
//                                }
//                                else
//                                {
//                                    right_valuation += Convert.ToInt32(digits);
//#if DEBUGCONSOLE
//                                    System.Console.Out.WriteLine("        Adding constant on the right: " + digits);
//#endif
//                                }
//                            }
//                                // Coeff + parameter
//                            else
//                            {
//                                // Get the parameter name
//                                string parameter = term.Substring(digits.Length);
//                                // Get the parameter valuation
//                                int value = EvaluateParameter(parameter, ParameterValuation);
//#if DEBUGCONSOLE
//                                System.Console.Out.WriteLine("        Found parameter:" + parameter + " = " + value);
//#endif

//                                // Update term
//                                int coeff = Convert.ToInt32(digits);
//                                if (left)
//                                {
//                                    left_valuation += coeff * value;
//#if DEBUGCONSOLE
//                                    System.Console.Out.WriteLine("        Adding on the left: " + coeff + "*" + value);
//#endif
//                                }
//                                else
//                                {
//                                    right_valuation += coeff * value;
//#if DEBUGCONSOLE
//                                    System.Console.Out.WriteLine("        Adding on the right: " + coeff + "*" + value);
//#endif
//                                }
//                            }
//                        }
//                            // Only a parameter
//                        else
//                        {
//                            // Get the parameter valuation
//                            int value = EvaluateParameter(term, ParameterValuation);
//#if DEBUGCONSOLE
//                            System.Console.Out.WriteLine("        Found parameter: " + term + " = " + value);
//#endif

//                            // Update term
//                            if (left)
//                            {
//                                left_valuation += value;
//#if DEBUGCONSOLE
//                                System.Console.Out.WriteLine("        Adding on the left: " + value);
//#endif
//                            }
//                            else
//                            {
//                                right_valuation += value;
//#if DEBUGCONSOLE
//                                System.Console.Out.WriteLine("        Adding on the right: " + value);
//#endif
//                            }
//                        }
//                    }
//                } // foreach term

//                // First: left, then right
//                if(left)
//                {
//                    left = false;
//                }
//            } // For each member

//            // Evaluate the inequality
//            switch (op)
//            {
//                case TimerOperationType.LessThanOrEqualTo:
//                    // Incompatible
//                    if (left_valuation > right_valuation)
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " > " + right_valuation);
//#endif
//                        // Add the negation of the inequality
//                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThan) + member_left);
//                    }
//                    // Compatible
//                    else
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " <= " + right_valuation);
//#endif
//                        return null;
//                    }
//                case TimerOperationType.LessThan:
//                    // Incompatible
//                    if (left_valuation >= right_valuation)
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " >= " + right_valuation);
//#endif
//                        // Add the negation of the inequality
//                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThanOrEqualTo) + member_left);
//                    }
//                    // Compatible
//                    else
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " < " + right_valuation);
//#endif
//                        return null;
//                    }
//                case TimerOperationType.Equals:
//                    // Incompatible
//                    if (left_valuation > right_valuation)
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Incompatible because " + left_valuation + " > " + right_valuation);
//#endif
//                        // Add the negation of the inequality
//                        return ConstraintOnParameters(member_right + opToString(TimerOperationType.LessThan) + member_left);
//                    }
//                    else if (right_valuation > left_valuation)
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Incompatible because " + right_valuation + " > " + left_valuation);
//#endif
//                        // Add the negation of the inequality
//                        return ConstraintOnParameters(member_left + opToString(TimerOperationType.LessThan) + member_right);
//                    }
//                    // Compatible
//                    else
//                    {
//#if DEBUGCONSOLE
//                        System.Console.Out.WriteLine("      Compatible because " + left_valuation + " = " + right_valuation);
//#endif
//                        return null;
//                    }
//            }
//            // This can't happen (from the parsing)
//            throw new Exception("The function should have returned already.");
//        }




        // Return a list of inequalities from the constraint that are incompatible with a reference valuation, and negate them
        // The constraint must be satisfiable
        // If the list is empty, the constraint is satisfied by the valuation
        //public List<Constraint> NegatedIncompatibleInequalities(Dictionary<string, int> ParameterValuation)
        //{
        //    List<Constraint> list = new List<Constraint>();

            // TO DO
            //// No false constraint allowed here
            //if (theConstraint == STRING_FALSE)
            //{
            //    throw new Exception("Attempting to test the compatibility of a false constraint with a reference valuation.");
            //}

            //// True constraint is satisfiable: empty list
            //if (theConstraint == STRING_TRUE)
            //{
            //    return list;
            //}

            //// Split between inequalities
            //string[] split = theConstraint.Split(new Char[] { '&' });

            //// Check and negate inequalities
            //foreach(string inequality in split)
            //{
            //    if (inequality != "")
            //    {
            //        // Negate the inequality
            //        Constraint negation = NegateInequalityIfIncompatible(inequality, ParameterValuation);
            //        // If null: already compatible
            //        if (negation != null)
            //        {
            //            list.Add(negation);
            //        }
            //    }
            //}
        //    return list;
        //}


        /**
         * See comments for PPL:Incompatible
        */
        public bool Incompatible(bool random_selection, int[] pi0, Constraint negated_inequality)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Entering Constraint:Incompatible()");
            System.Console.Out.WriteLine("negated_inequality = " + negated_inequality);
            System.Console.Out.WriteLine("pi0 = ");
            for (int i = 0; i < pi0.Length; i++)
            {
                System.Console.Out.WriteLine(i + " : " + pi0[i]);
            }
#endif
            Polyhedron resultPolyhedron = new Polyhedron();
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("resultPolyhedron before calling = " + resultPolyhedron.GetId());
#endif
            // Call the function on polyhedra (negated_inequality.thePolyhedron will be modified)
            bool result = thePolyhedron.Incompatible(random_selection, pi0, resultPolyhedron);
            // Update the negated inequality
            negated_inequality.thePolyhedron = resultPolyhedron;
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("resultPolyhedron after calling = " + resultPolyhedron.GetId());
            System.Console.Out.WriteLine("negated_inequality = " + negated_inequality);
            System.Console.Out.WriteLine("Exiting Constraint:Incompatible()");
#endif
            return result;
        }


        /************************************************************
        * OPERATIONS RETURNING A NEW OBJECT (no modification on this)
        ***********************************************************/

        // Convert a list of clock indexes into an int array, including the translation of the index for the constraint
        private static int[] IntArrayOfByteList(List<byte> l){
            int[] result = new int[l.Count];
            int index = 0;
            foreach (byte ClockId in l)
            {
                result[index] = ClockIdForConstraint(ClockId);
                index++;
            }
            return result;
        }

        // Remove the clocks not in "clocks"; keep the others and the parameters
        public Constraint KeepClocks(List<byte> ClocksToKeep)
        {
            // Test just in case; should not happen in normal mode, but may be called for exception and debug print
            if (!thePolyhedron.IsSatisfiable())
            {
                return Clone();
            }

            // Compute the list of clocks to remove
            List<byte> ClocksToRemove = ListDiff(ClockList, ClocksToKeep);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Keeping clocks {" + SomeClocksToString(ClocksToKeep)
                + "} i.e., hide {" + SomeClocksToString(ClocksToRemove)
                + "} in: \n" + this);
#endif

            // OPTIMIZATION: no hiding if has no clocks
            if (HasNoClocks())
            {
                Constraint clonedConstraint = Clone();

                // Update the list of clocks
                clonedConstraint.ClockList = new List<byte>(ClocksToKeep);

#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    Result: \n" + clonedConstraint);
                System.Console.Out.WriteLine("    No hiding because has no clocks ");
                System.Console.Out.WriteLine("    ====================");
#endif
                return clonedConstraint;
            }

            Constraint newConstraint = new Constraint();
            // Copy the polyhedron
            newConstraint.thePolyhedron = thePolyhedron.Clone();
            // Hide
            int[] ClocksToRemovePPL = IntArrayOfByteList(ClocksToRemove);
#if DEBUGCONSOLE
            string message = "    Calling PPL with {";
            for (int i = 0; i < ClocksToRemovePPL.Count(); i++)
            {
                message += ClocksToRemovePPL[i];
                if (i < ClocksToRemovePPL.Count() - 1)
                {
                    message += " , ";
                }
            }
            message += "} as argument.";
            System.Console.Out.WriteLine(message);
#endif
            newConstraint.thePolyhedron.Hide(ClocksToRemovePPL);

            // Update the list of clocks
            newConstraint.ClockList = new List<byte>(ClocksToKeep);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + newConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }


        // HIDE ALL CLOCKS IN THE CONSTRAINT (CLONE VERSION, this REMAINS UNCHANGED)
        public Constraint HideAllClocks()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    About to hide all clocks {" + ClocksToString() + "} in: \n" + this);
#endif

            // OPTIMIZATION: no hiding in true & no hiding of an empty set of clocks
            if (HasNoClocks())
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    (unchanged)");
                System.Console.Out.WriteLine("    ====================");
#endif
                return Clone();
            }

            // Copy the constraint
            Constraint newConstraint = Clone();
            //// Copy the polyhedron
            //newConstraint.thePolyhedron = thePolyhedron.Clone();
            // Hide
            newConstraint.thePolyhedron.Hide(IntArrayOfByteList(ClockList));

            // Empty the list of clocks
            newConstraint.ClockList = new List<byte>();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + newConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }


        /************************************************************
        * OPERATIONS MODIFYING THE OBJECT
        ***********************************************************/

        // Timed delay on the constraint
        public void Delay()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Delaying: \n" + this.ToString());
#endif
            if (HasNoClocks())
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    (unchanged)");
                System.Console.Out.WriteLine("    ====================");
#endif
                return; // OPTIMIZATION
            }
            thePolyhedron.TimeElapsing(IntArrayOfByteList(ClockList));

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }


        // Add a list of inequalities of the form 'x >= 0', for all 'x' in 'ClockList'
        public void EnsureClocksNonnegative(List<byte> ClockList)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Ensuring that clocks are non negative in: \n" + this);
#endif
            // Iterate on the clocks
            foreach (byte clockId in ClockList)
            {
                // Add the inequality
                thePolyhedron.AddGeqConst(ClockIdForConstraint(clockId), 0);
            }
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }


        // Test if all elements of l1 belong to l2
        private static bool ListIncluded(List<byte> l1, List<byte> l2){
            foreach (byte elem in l1)
            {
                if (!l2.Contains(elem)){
                    return false;
                }
            }
            return true;
        }



        // Computes the difference between l1 \ l2, i.e., the elements of l1 minus the elements of l2
        private static List<byte> ListDiff(List<byte> l1, List<byte> l2)
        {
            List<byte> toReturn = new List<byte>();
            foreach (byte elem in l1)
            {
                if (!l2.Contains(elem))
                {
                    toReturn.Add(elem);
                }
            }
            return toReturn;
        }



        // Conjunction between 2 constraints
        public void Conjunction(Constraint otherConstraint){
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    About to perform conjunction of \n" + this + "\n     with \n" + otherConstraint);
#endif
            // Verifying the equality of the clock list
            if (!(ListIncluded(ClockList, otherConstraint.ClockList)
                && ListIncluded(otherConstraint.ClockList, ClockList)))
            {
                throw new InvalidOperationException("Constraints " + ToString() +
                    " and " + otherConstraint.ToString() + " have different clock lists: {"
                    + ClocksToString() +
                    "} and {" + otherConstraint.ClocksToString() + "}.");
            }

            thePolyhedron.Intersection(otherConstraint.thePolyhedron);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }




        // Addition of a new clock set to 0 (if clockID already in, throws exception)
        public void AddClock(byte clockID)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding clock " + clockID + " to : \n" + this);
#endif

            // If the clock already exists, do nothing
            if (ClockList.Contains(clockID))
            {
                throw new InvalidOperationException("Attempt to add clock " + clockID.ToString()
                    + " to the list of clocks '" + ClocksToString() + "'.");
            }

            // Add the clock
            ClockList.Add(clockID);

            // Add the constraint 'clock = 0'
            thePolyhedron.AddEqConst(ClockIdForConstraint(clockID), 0);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result of addition: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }



        // Add an inequality of the form "clock ~ int"
        public void AddInequality(byte clockID, TimerOperationType op, int constant)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding inequality " + ClockToString(clockID) + opToString(op) + constant + " to : \n" + this);
#endif

            // Get the clock ID for the constraint
            byte UpdatedID = ClockIdForConstraint(clockID);
            switch (op)
            {
                case TimerOperationType.Equals:
                    thePolyhedron.AddEqConst(UpdatedID, constant);
                    break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    thePolyhedron.AddGeqConst(UpdatedID, constant);
                    break;
                case TimerOperationType.LessThanOrEqualTo:
                    thePolyhedron.AddLeqConst(UpdatedID, constant);
                    break;
                case TimerOperationType.GreaterThan:
                    thePolyhedron.AddGConst(UpdatedID, constant);
                    break;
                case TimerOperationType.LessThan:
                    thePolyhedron.AddLConst(UpdatedID, constant);
                    break;
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }



        // Add an inequality of the form "clock ~ parameter"
        public void AddParInequality(byte clockID, TimerOperationType op, string parameter)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Adding par inequality " + ClockToString(clockID) + opToString(op) + parameter + " to : \n" + this);
#endif

            // Get parameter name
            byte ParameterId = GetParameterId(parameter);
            // Get the clock ID for the constraint
            byte UpdatedID = ClockIdForConstraint(clockID);

            switch (op)
            {
                case TimerOperationType.Equals:
                    thePolyhedron.AddEqVar(UpdatedID, ParameterId);
                    break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    thePolyhedron.AddGeqVar(UpdatedID, ParameterId);
                    break;
                case TimerOperationType.LessThanOrEqualTo:
                    thePolyhedron.AddLeqVar(UpdatedID, ParameterId);
                    break;
                case TimerOperationType.GreaterThan:
                    thePolyhedron.AddGVar(UpdatedID, ParameterId);
                    break;
                case TimerOperationType.LessThan:
                    thePolyhedron.AddLVar(UpdatedID, ParameterId);
                    break;
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    ====================");
#endif
        }

        // Add a list of constraints
        // ETIENNE SAYS: there seem to be something strange here; contrarily to Conjunction, we do not test the list of clocks :(
        public void AddConstraints(List<Constraint> listOfConstraints)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    STARTING AddConstraints");
#endif
            // Iterate on the constraints
            foreach (Constraint otherConstraint in listOfConstraints)
            {
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("\n    ..... Adding inequality \n " + otherConstraint + "\n    ..... to : \n" + this);
                //System.Console.Out.WriteLine("\n    ..... this.thePolyhedron \n " + this.thePolyhedron.GetId());
                //System.Console.Out.WriteLine("\n    ..... otherConstraint.thePolyhedron \n " + otherConstraint.thePolyhedron.GetId());
#endif
                // Intersect
                thePolyhedron.Intersection(otherConstraint.thePolyhedron);
            }
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Result: \n" + this);
            System.Console.Out.WriteLine("    EXIT AddConstraints");
            System.Console.Out.WriteLine("    ====================");
#endif
        }

        /************************************************************
        * CONVERSION TO STRING
        ***********************************************************/
        

        public static string ClockToString(byte clockID)
        {
            return STRING_CLOCK_NAME + clockID.ToString();
        }


        public static string opToString(TimerOperationType op)
        {
            string result = "will be assigned later (function opToString)";
            switch (op)
            {
                case TimerOperationType.Equals:
                    result = "="; break;
                case TimerOperationType.GreaterThanOrEqualTo:
                    result = ">="; break;
                case TimerOperationType.LessThanOrEqualTo:
                    result = "<="; break;
                case TimerOperationType.GreaterThan:
                    result = ">"; break;
                case TimerOperationType.LessThan:
                    result = "<"; break;
            }
            return result;
        }


        private string ReplaceNamesIn(string theConstraint)
        {
            // First replace the var names with an independent var name
            // Pattern explanation:
            // - ([^A-Za-z_]|^) : one non letter and non '_' sign, or the beginning of the string
                // - STRING_VAR_NAME : the letter used to distinguish strings in PPL
            // - (\d+) : several digits
            //// - (\W|$) : one non letter, non '_' and non digit sign, or the end of the string
            // (note that it could react to v31blabla - but this would not happen, and I have nothing better to suggest)
            string pattern = @"([^A-Za-z_]|^)" + STRING_VAR_NAME + @"(\d+)"; // (\W|$)
            //theConstraint = "x23+x+bz32+x57+bx27+x31z<=p32+px21+x+x213+xp57+x317&x2<=x3&x2<=x1";
            string replacement = "$1" + STRING_VAR_NAME_TMP + "$2"; // $3
            Regex rgx = new Regex(pattern);
            string newConstraint = rgx.Replace(theConstraint, replacement);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    Replacement tmp : '" + theConstraint + "'  –>  '" + newConstraint + "'");
#endif

            // Replace each parameter by its name
            for (byte ParameterId = 0; ParameterId < NbParameters; ParameterId++)
            {
                // Retrieve name
                string ParameterName = ParameterNames[ParameterId];
                // Replace
                newConstraint = newConstraint.Replace(STRING_VAR_NAME_TMP + ParameterId, ParameterName);
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    After replacing parameters : '" + newConstraint + "'");
#endif

            // Useless to consider clocks if no clocks
            if (!HasNoClocks())
            {
                // Replace the clocks by their correct value
                foreach (byte ClockId in ClockList)
                {
                    // Get the id within the constraint
                    byte IdInConstraint = ClockIdForConstraint(ClockId);
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("      Id = '" + ClockId + "', i.e., '" + IdInConstraint + "' in constraint");
                    System.Console.Out.WriteLine("      Replacing '" + STRING_VAR_NAME_TMP + IdInConstraint + "' with '" + STRING_CLOCK_NAME + ClockId + "'");
#endif
                    // Replace the id in string, and update the clock name
                    newConstraint = newConstraint.Replace(STRING_VAR_NAME_TMP + IdInConstraint, ClockToString(ClockId));
                }

#if DEBUGCONSOLE
                System.Console.Out.WriteLine("    After replacing clocks : '" + newConstraint + "'");
#endif

            }

            // Return the constraint
            return newConstraint;
        }


        // Conversion to string
        public override string ToString()
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("BEGIN CONSTRAINT.TOSTRING");
#endif
            string theString = thePolyhedron.GetId();
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("thePolyhedron.GetId() = " + theString);
#endif

            // Add a mention if true
            // TO DO NonNegativeParametersConstraint.GetId() should be static
            if (theString == NonNegativeParametersConstraint.GetId())
            {
                theString += "\n (TRUE)";
            }

            // Replace names of clocks and parameters
            if (!thePolyhedron.IsTrue() && !thePolyhedron.IsFalse())
            {
                theString = ReplaceNamesIn(theString);
            }

            // DEBUGETIENNE
            // Add some blanks
            theString = theString.Replace(STRING_SEP_AND, "\n" + STRING_SEP_AND + " ");
            // Temp replace
            theString = theString.Replace("<=", "~");
            theString = theString.Replace("<", " < ");
            theString = theString.Replace("=", " = ");
            theString = theString.Replace("~", " <= ");
#if DEBUGCONSOLE
            theString += "\n(Clocks: " + ClocksToString() + ")";
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("END CONSTRAINT.TOSTRING");
#endif
            return "    " + theString;
        }



        // Convert a list of clocks to a string
        public static string SomeClocksToString(List<byte> ClockList)
        {
            // Case null
            if (ClockList == null) return "null";

            // Convert each clock to a string
            List<string> clocks_string = new List<string>(ClockList.Count());
            foreach (byte clockID in ClockList)
            {
                clocks_string.Add(ClockToString(clockID));
            }
            // Convert everything with a separator
            return string.Join(STRING_SEP_SET, clocks_string.ToArray());
        }



        // Convert all clocks to string
        public string ClocksToString()
        {
            return SomeClocksToString(ClockList);
        }


        // Return a unique identifier
        // Note that the constraints are under a canonical form
        // Clocks are "collapsed", so that constraint equivalent but with different clocks permutations still have the same id
        // (Hopefully!) Does not modify this!
        public string GetID(List<byte> OldClockList)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("\nCREATION OF CONSTRAINT ID");
#endif

            string Id = "";
            
            // Copy in order NOT to modify this.thePolyhedron
            Polyhedron polyhedronCloned = thePolyhedron.Clone();

            // Perform state space reduction by removing equivalent states (needs various costly operations)
            if (EQUIVALENCE_REDUCTION)
            {
                if (!HasNoClocks())
                {
                    byte MaxOldClockId = 0;

                    // 1) COMPUTE THE CORRESPONDENCE OLD -> NEW according to the OldClockList
                    Dictionary<byte, byte> correspondence = new Dictionary<byte, byte>();
                    byte CurrentNewId = FIRST_CLOCK;
                    foreach (byte OldClockId in OldClockList)
                    {
                        // Check if already captured (indeed, the list ClockList may contain the same clock several times!)
                        if (!correspondence.ContainsKey(OldClockId))
                        {
                            correspondence.Add(OldClockId, CurrentNewId);
                            // Update the new id for the next clock
                            CurrentNewId++;
                            // Update the max old id
                            if (OldClockId > MaxOldClockId)
                            {
                                MaxOldClockId = OldClockId;
                            }
                        }
                    }

                    //// 2) Check that all clocks are captured by the correspondence
                    //// ETIENNE SAYS: this is just a test, should be removed if functional!
                    //foreach (byte ClockId in ClockList)
                    //{
                    //    if (!correspondence.ContainsKey(ClockId))
                    //    {
                    //        //throw new Exception("Inconsistent clock lists in GetID()."
                    //        //    + "\n ClockList as a parameter: " + SomeClocksToString(OldClockList)
                    //        //    + "\n ClockList in Constraint: " + ClocksToString()
                    //        //    + "\n Constraint: " + theConstraint);
                    //    }
                    //}

                    // FIND THE MAX OLD CLOCK ID
                    foreach (byte OldClockId in ClockList)
                    {
                        // Update the max old id
                        if (OldClockId > MaxOldClockId)
                        {
                            MaxOldClockId = OldClockId;
                        }
                    }

    #if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  Correspondence OLD -> NEW:");
                    foreach (KeyValuePair<byte, byte> kvp in correspondence)
                    {
                        System.Console.Out.WriteLine("    " + kvp.Key + " –> " + kvp.Value);
                    }
    #endif

                    // 3) CREATE THE LIST OF NEW CLOCKS
                    foreach (byte clockID in OldClockList)
                    {
                        // Find the new value in the correspondence table
                        byte NewId = 0;
                        correspondence.TryGetValue(clockID, out NewId);
                        Id += NewId + Constants.SEPARATOR;
                    }

    #if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  Adding list of clocks to id: " + Id);
    #endif

                    // 4) RENAME ALL THE CLOCKS (not only the one in OldClockList), and create a bijection array
                    // Build the array old clock id -> new clock id
                    int[] renaming = new int[MaxOldClockId + 1 - FIRST_CLOCK];
                    // Also check if we have identity
                    bool identity = true;
                    for (byte OldClockId = FIRST_CLOCK; OldClockId <= MaxOldClockId; OldClockId++)
                    {
                        byte NewId;
                        // If in correspondence:
                        if (correspondence.ContainsKey(OldClockId))
                        {
                            // Get the new id
                            correspondence.TryGetValue(OldClockId, out NewId);
                        }
                        else
                        {
                            // Create a new id
                            NewId = CurrentNewId;
                            CurrentNewId++;
                        }
                        // Update the array
                        renaming[OldClockId - FIRST_CLOCK] = NewId - FIRST_CLOCK;
                        // Check identity (if still true)
                        if (identity && OldClockId != NewId)
                        {
                            identity = false;
                        }
                    }

    #if DEBUGCONSOLE
                    System.Console.Out.WriteLine("  Renaming array:");
                    for (int index = 0; index < renaming.Count(); index++)
                    {
                        System.Console.Out.WriteLine("    " + index + " –> " + renaming[index]);
                    }
    #endif

                    // Actual renaming (except if identity)
                    if (!identity)
                    {
    #if DEBUGCONSOLE
                        System.Console.Out.WriteLine("  Performing renaming");
    #endif
                        polyhedronCloned.Rename(NbParameters, renaming);
                    }
    #if DEBUGCONSOLE
                    else
                    {
                        System.Console.Out.WriteLine("  No renaming (identity)");
                    }
    #endif


                    // Put the constraint under normal form
                    // TO DO !! (simplify, remove dimensions, etc.) ?!
                }
            } // if EQUIVALENCE_REDUCTION

            // Get ID
            Id += polyhedronCloned.GetId();

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("END ID: " + Id);
#endif
            // 5) Return
            return Id;
        }


        /************************************************************
        * CLONE
        ***********************************************************/

        // Copy
        public Constraint Clone()
        {
            Constraint newConstraint = new Constraint();
            newConstraint.thePolyhedron = thePolyhedron.Clone();

            // Copy the list of clocks
            newConstraint.ClockList = new List<byte>(ClockList);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("    ====================");
            System.Console.Out.WriteLine("    Cloning: \n" + this);
            System.Console.Out.WriteLine("    Result: \n" + newConstraint);
            System.Console.Out.WriteLine("    ====================");
#endif
            return newConstraint;
        }



        private static void RunPPLTests()
        {
            Polyhedron p = new Polyhedron();
            Polyhedron p2 = new Polyhedron();

            System.Console.Out.WriteLine( "\n Id = " + p.GetId() + " ; Is satisfiable? " + p.IsSatisfiable());

            System.Console.Out.WriteLine("\n Id = " + p.GetId() + " ; Is true? " + p.IsTrue() + " ; Is false? " + p.IsFalse());

            p.AddLeqConst(2, 5);
            p.AddEqConst(3, 6);
            p.AddLConst(7, 1);
            System.Console.Out.WriteLine("\n Id = " + p.GetId() + " ; Is satisfiable? " + p.IsSatisfiable() + " ; Is true? " + p.IsTrue());

            p.AddLeqVar(1, 2);
            p.AddEqVar(4, 7);
            p.AddLVar(6, 10);
            System.Console.Out.WriteLine("\n Id = " + p.GetId() + " ; Is satisfiable? " + p.IsSatisfiable());

            p2 = p.Clone();
            System.Console.Out.WriteLine("\n Copy of p = " + p2.GetId());
            System.Console.Out.WriteLine("\n Modifying p = " + p.GetId());
            p.AddEqConst(1, 6);
            System.Console.Out.WriteLine("\n Id = " + p.GetId() + " ; Is satisfiable? " + p.IsSatisfiable());

            System.Console.Out.WriteLine("\n Copy: " + p2.GetId());

            p = new Polyhedron();
            p.AddEqConst(1, 6);
            System.Console.Out.WriteLine("\n First: " + p.GetId());
            p2 = new Polyhedron();
            p2.AddEqVar(2, 7);
            System.Console.Out.WriteLine("\n Second: " + p2.GetId());
            p.Intersection(p2);
            System.Console.Out.WriteLine("\n Intersection: " + p.GetId() + " (Other : " + p2.GetId() + ")");

            p2 = new Polyhedron();
            p2.AddEqConst(3, 6);
            System.Console.Out.WriteLine("\n First: " + p2.GetId());
            p = new Polyhedron();
            p.AddEqVar(2, 1);
            System.Console.Out.WriteLine("\n Second: " + p.GetId());
            p2.Intersection(p);
            System.Console.Out.WriteLine("\n (intersection done)");
            System.Console.Out.WriteLine("\n Intersection: " + p2.GetId() + " (Other : " + p.GetId() + ")");

            p = new Polyhedron();
            p.AddLeqVar(1, 2);
            p.AddLeqVar(2, 3);
            p.AddLeqVar(4, 5);
            p.AddLeqConst(5, 6);
            p.AddLeqConst(7, 2);
            int[] Vars = new int[] { 2, 5 };
            System.Console.Out.WriteLine("\n Before hiding: " + p.GetId());
            p.Hide(Vars);
            System.Console.Out.WriteLine("\n After hiding 2,5: " + p.GetId());

            p = new Polyhedron();
            p.AddGeqConst(1, 0);
            p.AddLConst(0, 5);
            p.AddLeqVar(0, 2);
            p.AddEqVar(0, 3);
            p.AddLVar(1, 4);
            p.AddLeqVar(4, 0);
//             0 <= u5
//& u3 < 5
//& u3 <= x0
//& u3 = x1
//& u5 < x2
//& x2 <= u3
            Vars = new int[] { };
            System.Console.Out.WriteLine("\n Before hiding: " + p.GetId());
            p.Hide(Vars);
            System.Console.Out.WriteLine("\n After hiding {}: " + p.GetId());
            Vars = new int[] { 3 };
            p.Hide(Vars);
            System.Console.Out.WriteLine("\n After hiding { 3 }: " + p.GetId());
            Vars = new int[] { 4 };
            p.Hide(Vars);
            System.Console.Out.WriteLine("\n After hiding { 4 }: " + p.GetId());



            p = new Polyhedron();
            p.AddLeqVar(1, 2);
            p.AddLeqVar(2, 3);
            p.AddLeqConst(7, 2);
            Vars = new int[] { 2, 5 };
            System.Console.Out.WriteLine("\n Before time elapsing: " + p.GetId());
            p.TimeElapsing(Vars);
            System.Console.Out.WriteLine("\n After time elapsing: " + p.GetId());

            p = new Polyhedron();
            p.AddLeqVar(1, 2);
            p.AddLeqVar(2, 3);
            p.AddLeqVar(4, 5);
            p.AddLeqConst(5, 6);
            p.AddLeqConst(7, 2);
            Vars = new int[] { 2, 5 };
            System.Console.Out.WriteLine("\n Before time elapsing: " + p.GetId());
            p.TimeElapsing(Vars);
            System.Console.Out.WriteLine("\n After time elapsing: " + p.GetId());

            p = new Polyhedron();
            p.AddLeqVar(0, 1);
            p.AddLeqVar(1, 2);
            Vars = new int[] { 1, 2, 0 };
            System.Console.Out.WriteLine("\n Before renaming: " + p.GetId());
            p.Rename(0, Vars);
            System.Console.Out.WriteLine("\n After renaming: " + p.GetId());

            p = new Polyhedron();
            p.AddLeqVar(0, 1);
            p.AddLeqVar(1, 2);
            Vars = new int[] { 1, 2, 5 };
            System.Console.Out.WriteLine("\n Before renaming: " + p.GetId());
            p.Rename(0, Vars);
            System.Console.Out.WriteLine("\n After renaming: " + p.GetId());

            p = new Polyhedron();
            p.AddLeqVar(0, 3);
            p.AddLeqVar(3, 5);
            Vars = new int[] { 0, 3, 4, 1, 5, 2};
            System.Console.Out.WriteLine("\n Before renaming: " + p.GetId());
            p.Rename(0, Vars);
            System.Console.Out.WriteLine("\n After renaming: " + p.GetId());

            // PI0-COMPATIBILITY
            bool was_random;
            Polyhedron result = new Polyhedron();
            int[] Pi0 = new int[] { 0, 1, 2, 3, 4, 5 };

            // Seeding
            Polyhedron.Seed();

            // true
            p = new Polyhedron();
            System.Console.Out.WriteLine("\n Polyhedron: " + p.GetId());
            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            // pi-incompatible
            p = new Polyhedron();
            p.AddLeqVar(0, 1);
            p.AddLVar(3, 2);
            p.AddEqConst(4, 5);
            System.Console.Out.WriteLine("\n Polyhedron: " + p.GetId());

            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible (no random selection)");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible (no random selection)");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            p = new Polyhedron();
            p.AddLeqVar(0, 1);
            p.AddLVar(3, 2);
            p.AddEqConst(4, 5);
            System.Console.Out.WriteLine("\n Polyhedron: " + p.GetId());

            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible (no random selection)");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible (no random selection)");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            // pi-compatible
            p = new Polyhedron();
            //result = new Polyhedron();
            p.AddLeqVar(0, 1);
            p.AddLVar(2, 3);
            p.AddLeqVar(4, 5);
            System.Console.Out.WriteLine("\n Polyhedron: " + p.GetId());
            was_random = p.Incompatible(false, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible (no random check)");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }
            was_random = p.Incompatible(true, Pi0, result);
            if (result.IsFalse())
            {
                System.Console.Out.WriteLine("Pi0-compatible");
                System.Console.Out.WriteLine("(Result: " + result.GetId() + ")");
            }
            else
            {
                System.Console.Out.WriteLine("Pi0-incompatible");
                System.Console.Out.WriteLine(result.GetId());
                System.Console.Out.WriteLine("Was random? " + was_random);
            }

            throw new Exception("Tests in progress. Please do not use this module now.");
        }




        public static void RunTests()
        {
            System.Console.Out.WriteLine("****************************************");
            System.Console.Out.WriteLine("Starting tests");
            System.Console.Out.WriteLine("****************************************");
            Constraint c1;
            Constraint c2;

            c1 = new Constraint();

            System.Console.Out.WriteLine("%%%%% C1: " + c1);

            c1.AddInequality(1, TimerOperationType.LessThan, 2);
            c1.AddInequality(2, TimerOperationType.GreaterThanOrEqualTo, 5);

            System.Console.Out.WriteLine("%%%%% C1 with 2 ineq: " + c1);

            c1 = constraintWithNonNegativeParameters();
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            System.Console.Out.WriteLine("%%%%% C1 init with 1 clock delayed: " + c1);
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            System.Console.Out.WriteLine("%%%%% C1 init with 2 clocks delayed: " + c1);
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            System.Console.Out.WriteLine("%%%%% C1 init with 3 clocks delayed: " + c1);
            System.Console.Out.WriteLine("%%%%% Id: " + c1.GetID(new List<byte>(new byte[] { 0, 1, 2 })));

            c1.AddInequality(1, TimerOperationType.LessThan, 5);
            c1.AddInequality(2, TimerOperationType.GreaterThanOrEqualTo, 2);

            System.Console.Out.WriteLine("%%%%% C1 with 2 ineq: " + c1);
            System.Console.Out.WriteLine("%%%%% Id 0 1 2: " + c1.GetID(new List<byte>(new byte[] { 0, 1, 2 })));
            System.Console.Out.WriteLine("%%%%% Id 0 0 1 2: " + c1.GetID(new List<byte>(new byte[] { 0, 0, 1, 2 })));
            System.Console.Out.WriteLine("%%%%% Id 2 0 1 1 2: " + c1.GetID(new List<byte>(new byte[] { 2, 0, 1, 1, 2 })));
            System.Console.Out.WriteLine("%%%%% Id 2 0: " + c1.GetID(new List<byte>(new byte[] { 2, 0 })));
            System.Console.Out.WriteLine("%%%%% Id 1: " + c1.GetID(new List<byte>(new byte[] { 1 })));
            System.Console.Out.WriteLine("%%%%% Id {}: " + c1.GetID(new List<byte>(new byte[] { })));

            c1.AddInequality(0, TimerOperationType.LessThanOrEqualTo, 1);
            System.Console.Out.WriteLine("%%%%% C1 false: " + c1);
            System.Console.Out.WriteLine("%%%%% Id: " + c1.GetID(new List<byte>(new byte[] { 0, 1, 2 })));

            c1 = constraintWithNonNegativeParameters();
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            c1.AddClock(c1.GetNewClockID());
            c1.Delay();
            c1.AddInequality(1, TimerOperationType.LessThan, 5);
            c1.AddParInequality(1, TimerOperationType.Equals, "u3");
            c1.AddParInequality(2, TimerOperationType.GreaterThan, "u5");

            System.Console.Out.WriteLine("%%%%% C1 with 3 par ineq: " + c1);
            System.Console.Out.WriteLine("%%%%% Id: " + c1.GetID(new List<byte>(new byte[] { 0, 1, 2 })));

            System.Console.Out.WriteLine("%%%%% C1 after keeping clocks {0,1,2}: " + c1.KeepClocks(new List<byte>(new byte[] { 1, 2, 0})));
            System.Console.Out.WriteLine("%%%%% C1 after keeping clocks {1,2}: " + c1.KeepClocks(new List<byte>(new byte[] { 1, 2 })));
            System.Console.Out.WriteLine("%%%%% C1 after keeping clocks {1}: " + c1.KeepClocks(new List<byte>(new byte[] { 1})));
            System.Console.Out.WriteLine("%%%%% C1 after keeping clocks {0}: " + c1.KeepClocks(new List<byte>(new byte[] { 0 })));
            System.Console.Out.WriteLine("%%%%% C1 projected onto P: " + c1.HideAllClocks());

            c1.AddInequality(0, TimerOperationType.GreaterThan, 1);
            System.Console.Out.WriteLine("%%%%% C1: " + c1);

            c2 = constraintWithNonNegativeParameters();
            c2.AddClock(c2.GetNewClockID());
            c2.Delay();
            c2.AddInequality(0, TimerOperationType.LessThan, 3);
            System.Console.Out.WriteLine("%%%%% C2: " + c2);

            c1.AddConstraints(new List<Constraint>(new Constraint[] { c2 }));
            System.Console.Out.WriteLine("%%%%% C1 intersected with c2: " + c1);



            System.Console.Out.WriteLine("****************************************");
            System.Console.Out.WriteLine("Ending tests");
            System.Console.Out.WriteLine("****************************************");
            throw new Exception("Tests in progress.");
        }


    }
}
