using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class EventPrefix : PARProcess
    {
        public bool IsUrgent;
        public Event Event;
        public PARProcess Process;

        public EventPrefix(Event e, PARProcess process, bool isUrgent)
        {
            Event = e;
            Process = process;
            IsUrgent = isUrgent;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(Event.GetID() + (IsUrgent ? Constants.EVENTPREFIX_URGENT : Constants.EVENTPREFIX) + Process.ProcessID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> list = new List<PARConfiguration>(1);

            string eventID = Event.GetEventID(eStep.GlobalEnv);
            string name = Event.GetEventName(eStep.GlobalEnv);
            list.Add(new PARConfiguration(Process, eventID, eventID != name ? name : null, eStep.GlobalEnv, false,
                                          IsUrgent, eStep.TheConstraint));
            return list;                
        }



        public override string ToString()
        {
            return "(" + Event + (IsUrgent? "->>": "->") + Process.ToString() + ")";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = Process.GetAlphabets(visitedDefinitionRefs);

            if (Event.ExpressionList != null)
            {
                foreach (Expression expression in Event.ExpressionList)
                {
                    if (expression.HasVar)
                    {
                        StringBuilder sb = new StringBuilder();
                        sb.AppendLine("ERROR - PAT FAILED to calculate the alphabet.");
                        sb.AppendLine("CAUSE - Event " + Event + " contains global variables!");
                        sb.AppendLine(
                            "REMEDY - 1) Avoid using global variables in events 2) Or manually specify the alphabet of the relevant process using the following syntax: \n\r\t #alphabet someProcess {X}; \n\rwhere X is a set of event names.");
                        throw new RuntimeException(sb.ToString());
                    }
                }
            }

            string name = Event.GetEventID(null);

            if (name != Constants.TAU && !list.Contains(name))
            {
                list.Add(name);
            }

            return list;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> toReturn = Process.GetGlobalVariables();

            if (Event.ExpressionList != null)
            {
                foreach (Expression expression in Event.ExpressionList)
                {
                    Common.Classes.Ultility.Ultility.AddList(toReturn, expression.GetVars());
                }
            }

            return toReturn;
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new EventPrefix(Event.ClearConstant(constMapping), Process.ClearConstant(constMapping), IsUrgent);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override void CalculateIdleTime(RTSConfiguration eStep, DBM delayDBM)
        //{
        //    if (!IsUrgent)
        //    {
        //        if (Event.GetEventName(eStep.GlobalEnv) != Constants.TAU || !RTSAssertionRefinementTime.TauUrgency)
        //        {
        //            return;
        //        }
        //    }

        //    //return eStep.GlobalEnv.DBM;
        //    DBM newDBM = eStep.DBM.Clone();
        //    delayDBM.TimerArray = newDBM.TimerArray;
        //    delayDBM.Matrix = newDBM.Matrix;
        //    delayDBM.IsCanonicalForm = newDBM.IsCanonicalForm;
        //}
    
        //public override bool IsTime()
        //{
        //    return Process.IsTime();
        //}
    }
}