using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class Hiding : PARProcess
    {
        public  PARProcess Process;
        public  EventCollection HidingAlphabets;

        public Hiding(PARProcess process, EventCollection alphabets)
        {
            if (process is Hiding)
            {
                Process = (process as Hiding).Process;
                List<Event> events = new List<Event>((process as Hiding).HidingAlphabets);
                
                foreach (Event item in alphabets)
                {
                    if (!events.Contains(item))
                    {
                        events.Add(item);
                    }
                }

                HidingAlphabets = new EventCollection(events);        
            }
            else {
                Process = process;
                HidingAlphabets = alphabets;
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.HIDING + Process.ProcessID + Constants.SEPARATOR + HidingAlphabets.ProcessID);
        }




        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnlist = Process.MoveOneStep(eStep);

            foreach (PARConfiguration step in returnlist)
            {
                step.Process = new Hiding(step.Process, HidingAlphabets);
                if (HidingAlphabets.ContainEventName(step.Event))
                {
                    step.DisplayName = "[" + step.Event + "]";
                    step.Event = Constants.TAU;
                }
            }

            //int i = 0;

            //while (i < returnlist.Count)
            //{
            //    PARConfiguration step = returnlist[i];

            //    if (HidingAlphabets.ContainEventName(step.Event))
            //    {
            //        step.HidenEvent = "[" + step.Event + "]";
            //        step.Event = Constants.TAU;

            //        if (PARConfiguration.TauUrgency)
            //        {
            //            Constraint newConstraint = step.TheConstraint.Clone();
            //            newConstraint.AddInequality(newClock, TimerOperationType.Equals, 0);

            //            if (newConstraint.IsSatisfiable())
            //            {
            //                step.Process = new Hiding(step.Process, HidingAlphabets);
            //                i++;
            //                idleConstraint.AddInequality(newClock, TimerOperationType.Equals, 0);
            //            }
            //            else
            //            {
            //                returnlist.RemoveAt(i);
            //            }
            //        }
            //        else
            //        {
            //            step.Process = new Hiding(step.Process, HidingAlphabets);
            //            i++;
            //        }

            //    }
            //    else
            //    {
            //        step.Process = new Hiding(step.Process, HidingAlphabets);
            //        i++;
            //    }
            //}


            return returnlist;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> returnlist = Process.SyncOutput(eStep);
            for (int i = 0; i < returnlist.Count; i++)
            {
                returnlist[i].Process = new Hiding(returnlist[i].Process, HidingAlphabets);
            }

            return returnlist;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> returnlist = Process.SyncInput(eStep);
            for (int i = 0; i < returnlist.Count; i++)
            {
                returnlist[i].Process = new Hiding(returnlist[i].Process, HidingAlphabets);
            }

            return returnlist;
        }

        public override string ToString()
        {
            return "(" + Process + " \\ {" + Common.Classes.Ultility.Ultility.PPStringList(HidingAlphabets) + "})";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> returnlist = Process.GetAlphabets(visitedDefinitionRefs);
            foreach (string alphabet in HidingAlphabets.EventNames)
            {
                returnlist.Remove(alphabet);
            }   

            //int counter = 0;
            //while (counter < returnlist.Count)
            //{
            //    string step = returnlist[counter];
            //    if (HidingAlphabets.ContainEventName(step))
            //    {
            //        returnlist.Remove(step);
            //    }
            //    else
            //    {
            //        counter++;
            //    }
            //}
            return returnlist;
        }

        public override List<string> GetGlobalVariables()
        {
            return Process.GetGlobalVariables();
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new Hiding(Process.ClearConstant(constMapping), HidingAlphabets.ClearConstant(constMapping));
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return Process.GetTopLevelConcurrency(visitedDef);
        }

        //public override List<byte> GetActiveTimer()
        //{
        //    return Process.GetActiveTimer();
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = Process.SetTimer(timerID, clocks);
            if (process == null)
            {
                return null;
            }

            return new Hiding(process, HidingAlphabets);
        }

        //public override bool IsTime()
        //{
        //    return Process.IsTime();
        //}
    }
}