using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexExternalChoice : PARProcess
    {
        public List<PARProcess> Processes;

        //interleave based on the indexed definitions.
        public IndexedProcess IndexedProcessDefinition;

        public IndexExternalChoice(IndexedProcess definition)
        {
            IndexedProcessDefinition = definition;
        }

        public IndexExternalChoice(List<PARProcess> processes)
        {
            Processes = new List<PARProcess>();
            SortedDictionary<string, bool> existed = new SortedDictionary<string, bool>();
            //IdleTimeConstraints = new List<ClockConstraints>();

            foreach (PARProcess proc in processes)
            {
                if (proc is IndexExternalChoice && (proc as IndexExternalChoice).Processes != null)
                {
                    List<PARProcess> processes1 = (proc as IndexExternalChoice).Processes;

                    foreach (PARProcess processe in processes1)
                    {
                        if (!existed.ContainsKey(processe.ProcessID))
                        {
                            Processes.Add(processe);
                            //IdleTimeConstraints.AddRange(processe.IdleTimeConstraints);
                            existed.Add(processe.ProcessID, false);
                        }
                    }
                }
                else
                {
                    if (!existed.ContainsKey(proc.ProcessID))
                    {
                        Processes.Add(proc);
                        //IdleTimeConstraints.AddRange(proc.IdleTimeConstraints);
                        existed.Add(proc.ProcessID, false);
                    }
                }
            }

            Debug.Assert(Processes.Count > 0);

            if (Processes.Count > 1)
            {
                StringBuilder ID = new StringBuilder();
                foreach (string id in existed.Keys)
                {
                    ID.Append(Constants.EXTERNAL_CHOICE);
                    ID.Append(id);
                }
                ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
            }
            else
            {
                ProcessID = Processes[0].ProcessID;
            }
        }
        



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                List<PARConfiguration> lists = Processes[i].MoveOneStep(eStep);

                for (int l = 0; l < lists.Count; l++)
                {
                    PARConfiguration configuration = lists[l];
                    if (configuration.Event == Constants.TAU)
                    {
                        List<PARProcess> newProcess = new List<PARProcess>(Processes);
                        newProcess[i] = configuration.Process;
                        IndexExternalChoice choice = new IndexExternalChoice(newProcess);
                        configuration.Process = choice;
                    }

                    toReturn.Add(configuration);
                }
            }

            return toReturn;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> toReturn = new List<PARConfigurationWithChannelData>();

            for (int i = 0; i < Processes.Count; i++)
            {
                toReturn.AddRange(Processes[i].SyncOutput(eStep));
            }

            return toReturn;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                toReturn.AddRange(Processes[i].SyncInput(eStep));
            }

            return toReturn;
        }

        public override string ToString()
        {
            if (Processes == null)
            {
                return "[*]" + IndexedProcessDefinition;
            }

            string result = "(" + Processes[0];
            for (int i = 1; i < Processes.Count; i++)
            {
                result += "[*]";
                result += Processes[i].ToString();
            }
            result += ")";
            return result;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> list = new HashSet<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                list.UnionWith(Processes[i].GetAlphabets(visitedDefinitionRefs));
            }
            return list;
        }

        public override List<string> GetGlobalVariables()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetGlobalVariables();
            }

            List<string> Variables = new List<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                List<string> temp = Processes[i].GetGlobalVariables();
                Common.Classes.Ultility.Ultility.AddList(Variables, temp);
            }
            return Variables;
        }

        public override List<string> GetChannels()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetChannels();
            }

            List<string> channels = new List<string>();
            for (int i = 0; i < Processes.Count; i++)
            {
                List<string> temp = Processes[i].GetChannels();
                Common.Classes.Ultility.Ultility.AddList<string>(channels, temp);
            }
            return channels;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            List<PARProcess> newnewListProcess = Processes;
            if (Processes == null)
            {
                if (Specification.IsParsing)
                {
                    return new IndexExternalChoice(IndexedProcessDefinition.ClearConstant(constMapping));
                }

                //return new IndexChoice(IndexedProcessDefinition.GetIndexedProcesses(constMapping));
                newnewListProcess = IndexedProcessDefinition.GetIndexedProcesses(constMapping);
            }

            List<PARProcess> newListProcess = new List<PARProcess>();
            for (int i = 0; i < newnewListProcess.Count; i++)
            {
                PARProcess newProc = newnewListProcess[i].ClearConstant(constMapping);
                if (!(newProc is Stop))
                {
                    newListProcess.Add(newProc);
                }
            }

            //all processes are stop processes
            if (newListProcess.Count == 0)
            {
                return new Stop();
            }

            return new IndexExternalChoice(newListProcess);
        }

        public override bool MustBeAbstracted()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.MustBeAbstracted();
            }

            for (int i = 0; i < Processes.Count; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }

            return false;
        }

        //public override void CalculateIdleTime(RTSConfiguration eStep, DBM delayDBM)
        //{
        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Processes[i].CalculateIdleTime(eStep, delayDBM);
        //    }
        //}

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> activeTimers = new List<byte>();

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Common.Classes.Ultility.Ultility.Union(activeTimers, Processes[i].GetActiveTimer());
        //    }
        //    return activeTimers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            List<PARProcess> newListProcess = new List<PARProcess>();
            bool anyNeedtimer = false;

            for (int i = 0; i < this.Processes.Count; i++)
            {
                PARProcess newProc = Processes[i].SetTimer(timerID, clocks);
                if (newProc == null)
                {
                    newListProcess.Add(Processes[i]);
                }
                else
                {
                    anyNeedtimer = true;
                    newListProcess.Add(newProc);
                }
            }

            if (anyNeedtimer)
            {
                return new IndexExternalChoice(newListProcess);
            }

            return null;
        }

        //public override bool IsTime()
        //{
        //    if (Processes == null)
        //    {
        //        return IndexedProcessDefinition.Process.IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }
        //    return false;
        //}
    }
}