using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class IndexInterleave : PARProcess
    {
        public List<PARProcess> Processes;

        public IndexedProcess IndexedProcessDefinition;

        public IndexInterleave(IndexedProcess definition)
        {
            IndexedProcessDefinition = definition;
        }

        public IndexInterleave(List<PARProcess> processes)
        {
            Processes = processes;

            string ID = Processes[0].ProcessID;

            for (int i = 1; i < Processes.Count; i++)
            {
                ID += Constants.INTERLEAVE + Processes[i].ProcessID;
            }

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();
            List<List<PARConfiguration>> terminateSteps = new List<List<PARConfiguration>>();
            //bool allReadyToTerminate = true;

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfiguration> list1 = process.MoveOneStep(eStep);
                //bool readyToTerminate = false;
                terminateSteps.Add(new List<PARConfiguration>());

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];

                    if (step.Event != Constants.TERMINATION)
                    {
                        //otherwise, the participating process is process-i only.
                        if (AssertionBase.CalculateParticipatingProcess)
                        {
                            step.ParticipatingProcesses = new string[] { i.ToString() };
                        }

                        List<PARProcess> newProcess = new List<PARProcess>(Processes);
                        newProcess[i] = step.Process;
                        IndexInterleave interleave = new IndexInterleave(newProcess);
                        step.Process = interleave;

                        returnList.Add(step);
                    }
                    else
                    {
                        //readyToTerminate = true;
                        terminateSteps[i].Add(step);
                    }
                }
                //}
                //allReadyToTerminate = allReadyToTerminate && readyToTerminate;

                //to check whether there are synchoronous channel input/output
                if (SpecificationBase.HasSyncrhonousChannel)
                {
                    SynchronousChannelInputOutput(returnList, i, eStep, null);
                }
            }

            //the following is for termination only.
            bool allReadyToTerminate = true;
            foreach (List<PARConfiguration> rtsConfigurations in terminateSteps)
            {
                if (rtsConfigurations.Count == 0)
                {
                    allReadyToTerminate = false;
                    break;
                }
            }

            if (allReadyToTerminate)
            {
                returnList.AddRange(TerminationSteps(eStep, terminateSteps));
                //idleDBM.AddConstraint(newClock, TimerOperationType.Equals, 0);
            }

            return returnList;
        }



        private void SynchronousChannelInputOutput(List<PARConfiguration> returnList, int i, PARConfiguration step, string evt)
        {
            List<PARConfigurationWithChannelData> outputs = Processes[i].SyncOutput(step);

            foreach (PARConfigurationWithChannelData vm in outputs)
            {
                if (evt != null && vm.Event != evt)
                {
                    continue;
                }

                PARProcess output = vm.Process;

                for (int k = 0; k < Processes.Count; k++)
                {
                    if (k != i)
                    {
                        List<PARConfiguration> syncedProcess = Processes[k].SyncInput(vm);

                        foreach (PARConfiguration p in syncedProcess)
                        {
                            List<PARProcess> newProcess = new List<PARProcess>(Processes);
                            newProcess[i] = output;
                            newProcess[k] = p.Process;

                            IndexInterleave interleave = new IndexInterleave(newProcess);

                            Constraint newConstraint = vm.TheConstraint.Clone();
                            newConstraint.Conjunction(p.TheConstraint);

                            PARConfiguration newStep = new PARConfiguration(interleave, vm.Event, vm.DisplayName, vm.GlobalEnv, false, vm.IsUrgent || p.IsUrgent, newConstraint);
                            newStep.IsAtomic = vm.IsAtomic || p.IsAtomic;

                            if (AssertionBase.CalculateParticipatingProcess)
                            {
                                newStep.ParticipatingProcesses = new string[] { i.ToString(), k.ToString() };
                            }

                            returnList.Add(newStep);
                        }
                    }
                }
            }
        }

        private List<PARConfiguration> TerminationSteps(PARConfiguration eStep, List<List<PARConfiguration>> terminateSteps)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();
            List<bool> isAtomic = new List<bool>();

            //maps an event to the list of resulting processes.
            isAtomic.Add(false);

            List<List<Constraint>> syncedConstraints = new List<List<Constraint>>();
            syncedConstraints.Add(new List<Constraint>());

            for (int i = 0; i < Processes.Count; i++)
            {
                List<PARConfiguration> steps = terminateSteps[i];

                foreach (PARConfiguration step in steps)
                {
                    if (syncedConstraints[0].Count == i)
                    {
                        for (int j = 0; j < syncedConstraints.Count; j++)
                        {
                            syncedConstraints[j].Add(step.TheConstraint);

                            if (step.IsAtomic)
                            {
                                for (int k = 0; k < isAtomic.Count; k++)
                                {
                                    isAtomic[k] = true;
                                }
                            }
                        }
                    }
                    else
                    {
                        //If there non-determinism, clone and then add.
                        List<List<Constraint>> toAddConstraint = new List<List<Constraint>>(syncedConstraints.Count);

                        for (int j = 0; j < syncedConstraints.Count; j++)
                        {
                            List<Constraint> newConstraintList = new List<Constraint>();
                            List<Constraint> Constraintlist = syncedConstraints[j];

                            for (int k = 0; k < Constraintlist.Count - 1; k++)
                            {
                                newConstraintList.Add(Constraintlist[k]);
                            }

                            newConstraintList.Add(step.TheConstraint);
                            toAddConstraint.Add(newConstraintList);
                            isAtomic.Add(step.IsAtomic);
                        }

                        syncedConstraints.AddRange(toAddConstraint);
                    }
                }
            }

            for (int i = 0; i < syncedConstraints.Count; i++)
            {
                Constraint newconstraint = Constraint.constraintWithNonNegativeParameters(); // trueConstraint
                bool isAtomicBool = false;
                for (int k = 0; k < syncedConstraints[i].Count; k++)
                {
                    newconstraint.Conjunction(syncedConstraints[i][k]);
                    isAtomicBool = isAtomicBool || isAtomic[i];
                }

                PARConfiguration tmpStep = new PARConfiguration(new Stop(), Constants.TERMINATION, null, eStep.GlobalEnv, false, false, newconstraint);
                tmpStep.IsAtomic = isAtomicBool;

                if (AssertionBase.CalculateParticipatingProcess)
                {
                    tmpStep.ParticipatingProcesses = new string[Processes.Count];
                    for (int j = 0; j < Processes.Count; j++)
                    {
                        tmpStep.ParticipatingProcesses[j] = j.ToString();
                    }
                }

                returnList.Add(tmpStep);
            }

            return returnList;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> returnList = new List<PARConfigurationWithChannelData>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfigurationWithChannelData> list1 = process.SyncOutput(eStep);

                for (int j = 0; j < list1.Count; j++)
                {
                    PARConfiguration step = list1[j];

                    List<PARProcess> newProcess = new List<PARProcess>(Processes.Count);
                    newProcess.AddRange(Processes);
                    newProcess[i] = step.Process;

                    step.Process = new IndexInterleave(newProcess);
                    returnList.Add(list1[j]);
                }
            }

            return returnList;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> returnList = new List<PARConfiguration>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                List<PARConfiguration> list1 = process.SyncInput(eStep);

                if (list1.Count > 0)
                {
                    for (int j = 0; j < list1.Count; j++)
                    {
                        PARConfiguration step = list1[j];

                        List<PARProcess> newProcess = new List<PARProcess>(Processes.Count);
                        newProcess.AddRange(Processes);
                        newProcess[i] = step.Process;

                        IndexInterleave interleave = new IndexInterleave(newProcess);
                        step.Process = interleave;
                        returnList.Add(step);
                    }
                }
            }

            return returnList;
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetAlphabets(visitedDefinitionRefs);
            }

            HashSet<string> toReturn = new HashSet<string>();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                toReturn.UnionWith(process.GetAlphabets(visitedDefinitionRefs));
            }

            return toReturn;
        }

        public override List<string> GetGlobalVariables()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetGlobalVariables();
            }

            List<string> Variables = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(Variables, var.GetGlobalVariables());
            }

            return Variables;
        }

        public override List<string> GetChannels()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.GetChannels();
            }

            List<string> channels = new List<string>();

            foreach (PARProcess var in Processes)
            {
                Common.Classes.Ultility.Ultility.AddList(channels, var.GetChannels());
            }

            return channels;
        }

        public override string ToString()
        {
            if (Processes == null)
            {
                return "|||" + IndexedProcessDefinition;
            }

            StringBuilder s = new StringBuilder();

            for (int i = 0; i < Processes.Count; i++)
            {
                PARProcess process = Processes[i];
                s.Append(process + "\r\n|||");
            }
            return "(" + s.ToString().TrimEnd('|') + ")";
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            List<PARProcess> newProceses = new List<PARProcess>();

            if (Specification.IsParsing)
            {
                if (Processes == null)
                {
                    return new IndexInterleave(IndexedProcessDefinition.ClearConstant(constMapping));
                }

                foreach (PARProcess procese in Processes)
                {
                    newProceses.Add(procese.ClearConstant(constMapping));
                }

                return new IndexInterleave(newProceses);
            }

            List<PARProcess> newnewProcesses = Processes;

            if (Processes == null)
            {
                newnewProcesses = IndexedProcessDefinition.GetIndexedProcesses(constMapping);
            }

            int size = newnewProcesses.Count;
            Dictionary<string, int> processCounters = new Dictionary<string, int>(size);

            for (int i = 0; i < size; i++)
            {
                PARProcess newProc = newnewProcesses[i].ClearConstant(constMapping);

                if (newProc is IndexInterleave)
                {
                    List<PARProcess> processes = (newProc as IndexInterleave).Processes;

                    foreach (var processe in processes)
                    {
                        string tmp = processe.ProcessID;
                        if (!processCounters.ContainsKey(tmp))
                        {
                            newProceses.Add(processe);
                            processCounters.Add(tmp, 1);
                        }
                        else
                        {
                            processCounters[tmp] = processCounters[tmp] + 1;
                        }
                    }
                }
                else if (newProc is IndexInterleaveAbstract)
                {
                    IndexInterleaveAbstract intAbs = newProc as IndexInterleaveAbstract;

                    foreach (PARProcess processe in intAbs.Processes)
                    {
                        string tmp = processe.ProcessID;
                        if (!processCounters.ContainsKey(tmp))
                        {
                            newProceses.Add(processe);
                            processCounters.Add(tmp, intAbs.ProcessesActualSize[tmp]);
                        }
                        else
                        {
                            processCounters[tmp] = processCounters[tmp] + intAbs.ProcessesCounter[tmp];
                        }
                    }
                }
                else
                {
                    if (!processCounters.ContainsKey(newProc.ProcessID))
                    {
                        newProceses.Add(newProc);
                        processCounters.Add(newProc.ProcessID, 1);
                    }
                    else
                    {
                        processCounters[newProc.ProcessID] = processCounters[newProc.ProcessID] + 1;
                    }
                }
            }

            foreach (KeyValuePair<string, int> pair in processCounters)
            {
                if (pair.Value > 1 || pair.Value == -1)
                {
                    IndexInterleaveAbstract toReturn = new IndexInterleaveAbstract(newProceses, processCounters);
                    toReturn.ProcessesActualSize = new Dictionary<string, int>(processCounters);
                    return toReturn;
                }
            }

            return new IndexInterleave(newProceses);
        }

        public override bool MustBeAbstracted()
        {
            if (Processes == null)
            {
                return IndexedProcessDefinition.Process.MustBeAbstracted();
            }


            for (int i = 0; i < Processes.Count; i++)
            {
                if (Processes[i].MustBeAbstracted())
                {
                    return true;
                }
            }
            return false;
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return this;
        }      

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> activeTimers = new List<byte>();

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        Common.Classes.Ultility.Ultility.Union(activeTimers, Processes[i].GetActiveTimer());//eStep));
        //    }
        //    return activeTimers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            List<PARProcess> newListProcess = new List<PARProcess>();
            bool anyNeedtimer = false;

            for (int i = 0; i < this.Processes.Count; i++)
            {
                PARProcess newProc = Processes[i].SetTimer(timerID, clocks);
                if (newProc == null)
                {
                    newListProcess.Add(Processes[i]);
                }
                else
                {
                    anyNeedtimer = true;
                    newListProcess.Add(newProc);
                }
            }

            //all processes are stop processes
            if (anyNeedtimer)
            {
                IndexInterleave index = new IndexInterleave(newListProcess);
                return index;
            }

            return null;
        }

        //public override bool IsTime()
        //{
        //    if (Processes == null)
        //    {
        //        return IndexedProcessDefinition.Process.IsTime();
        //    }

        //    for (int i = 0; i < Processes.Count; i++)
        //    {
        //        if (Processes[i].IsTime())
        //        {
        //            return true;
        //        }
        //    }
        //    return false;
        //}
   }
}