﻿using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;

namespace PAT.PAR.LTS
{
    public class IndexedProcess
    {
        public PARProcess Process;
        public List<ParallelDefinition> Definitions;

        public IndexedProcess(PARProcess process, List<ParallelDefinition> definitions)
        {
            Process = process;
            Definitions = definitions;
        }

        public List<string> GetGlobalVariables(List<string> visitedDef)
        {
            List<string> Variables = new List<string>();

            foreach (ParallelDefinition definition in Definitions)
            {
                Common.Classes.Ultility.Ultility.AddList(Variables, definition.GetGlobalVariables());
            }

            Common.Classes.Ultility.Ultility.AddList(Variables, Process.GetGlobalVariables());
            return Variables;
        }

        public void GlobalVariablesAsIndex(List<string> visitedDef)
        {
            foreach (ParallelDefinition definition in Definitions)
            {
                if(definition.GetGlobalVariables().Count > 0)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.AppendLine("ERROR - PAT FAILED to calculate the alphabet of process " + this.ToString() + ".");
                    sb.AppendLine("CAUSE - The process definition contains global variables!");
                    sb.AppendLine("REMEDY - 1) Avoid using global variables in process 2) Or manually specify the alphabet of the relevant process using the following syntax: \n\r\t #alphabet someProcess {X}; \n\rwhere X is a set of event names.");
                    throw new RuntimeException(sb.ToString());
                }
            }
        }

        public List<PARProcess> GetIndexedProcesses(Dictionary<string, Expression> constMapping)
        {
            List<ParallelDefinition> newDefinitions = new List<ParallelDefinition>();
            foreach (ParallelDefinition definition in Definitions)
            {
                ParallelDefinition newPD = definition.ClearConstant(constMapping);
                newDefinitions.Add(newPD);
            }

            List<PARProcess> processes = new List<PARProcess>(16);

            foreach (ParallelDefinition pd in newDefinitions)
            {
                pd.DomainValues.Sort();
            }

            List<List<Expression>> list = new List<List<Expression>>();
            foreach (int v in newDefinitions[0].DomainValues)
            {
                List<Expression> l = new List<Expression>(newDefinitions.Count);
                l.Add(new IntConstant(v));
                list.Add(l);
            }

            for (int i = 1; i < newDefinitions.Count; i++)
            {
                List<List<Expression>> newList = new List<List<Expression>>();
                List<int> domain = newDefinitions[i].DomainValues;

                for (int j = 0; j < list.Count; j++)
                {
                    foreach (int i1 in domain)
                    {
                        List<Expression> cList = new List<Expression>(list[j]);
                        cList.Add(new IntConstant(i1));
                        newList.Add(cList);
                    }
                }
                list = newList;
            }

            foreach (List<Expression> constants in list)
            {
                //Dictionary<string, Expression> constMappingNew = new Dictionary<string, Expression>(constMapping);
                Dictionary<string, Expression> constMappingNew = new Dictionary<string, Expression>();
                for (int i = 0; i < constants.Count; i++)
                {
                    Expression constant = constants[i];
                    //constant.BuildVars();
                    constMappingNew.Add(newDefinitions[i].Parameter, constant);

                }

                PARProcess newProcess = Process.ClearConstant(constMappingNew);
                processes.Add(newProcess);
            }

            return processes;
        }

        public IndexedProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            PARProcess newProcess = Process.ClearConstant(constMapping);

            List<ParallelDefinition> newDefinitions = new List<ParallelDefinition>(Definitions.Count);

            foreach (ParallelDefinition definition in Definitions)
            {
                newDefinitions.Add(definition.ClearConstant(constMapping));
            }

            return new IndexedProcess(newProcess, newDefinitions);
        }

        public override string ToString()
        {
            string returnString = "";
            foreach (ParallelDefinition list in Definitions)
            {
                returnString += list.ToString() + ";";
            }

            return returnString.TrimEnd(';') + "@" + Process;
        }
    }
}
