using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class Interrupt : PARProcess
    {
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;

        public Interrupt(PARProcess firstProcess, PARProcess secondProcess)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.INTERRUPT + SecondProcess.ProcessID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> list1 = FirstProcess.MoveOneStep(eStep);

            for (int i = 0; i < list1.Count; i++)
            {
                PARConfiguration step = list1[i];
                if (step.Event != Constants.TERMINATION)
                {
                    Interrupt inter = new Interrupt(step.Process, SecondProcess);
                    step.Process = inter;
                }
            }

            List<PARConfiguration> list2 = SecondProcess.MoveOneStep(eStep);
            for (int i = 0; i < list2.Count; i++)
            {
                PARConfiguration step = list2[i];

                if (step.Event == Constants.TAU)
                {
                    Interrupt inter = new Interrupt(FirstProcess, step.Process);

                    step.Process = inter;
                }

                list1.Add(step);
            }

            return list1;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> list1 = FirstProcess.SyncOutput(eStep);

            for (int i = 0; i < list1.Count; i++)
            {
                PARConfiguration step = list1[i];
                if (step.Event != Constants.TERMINATION)
                {
                    Interrupt inter = new Interrupt(step.Process, SecondProcess);
                    step.Process = inter;
                }
            }

            list1.AddRange(SecondProcess.SyncOutput(eStep));
            //List<RTSConfigurationWithChannelData> list2 = SecondProcess.SyncOutput(eStep);
            //for (int i = 0; i < list2.Count; i++)
            //{
            //    RTSConfiguration step = list2[i];

                //DBM newDBM = step.DBM.Clone();
                //newDBM.Conjunction(firstIdle);
                ////foreach (ClockConstraints list in FirstProcess.IdleTimeConstraints)
                ////{
                ////    newDBM.AddConstraint(list.TimeID, list.Operator, list.Constant);
                ////}
                //step.DBM = newDBM;

            //    list1.Add(list2[i]);               
            //});

            return list1;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> list1 = FirstProcess.SyncInput(eStep);

            //DBM firstIdle = eStep.InitialDBM.Delay();
            //DBM secondIdle = firstIdle.Clone();

            //FirstProcess.CalculateIdleTime(eStep, firstIdle);
            //SecondProcess.CalculateIdleTime(eStep, secondIdle);

            for (int i = 0; i < list1.Count; i++)
            {
                PARConfiguration step = list1[i];
                if (step.Event != Constants.TERMINATION)
                {
                    Interrupt inter = new Interrupt(step.Process, SecondProcess);
                    step.Process = inter;
                }

                //DBM newDBM = step.DBM.Clone();
                //newDBM.Conjunction(secondIdle);
                //foreach (ClockConstraints list in SecondProcess.IdleTimeConstraints)
                //{
                //    newDBM.AddConstraint(list.TimeID, list.Operator, list.Constant);
                //}
                //step.DBM = newDBM;
                
                //returnList.Add(step);
            }

            list1.AddRange(SecondProcess.SyncInput(eStep));

            //List<RTSConfiguration> list2 = SecondProcess.SyncInput(eStep);
            //for (int i = 0; i < list2.Count; i++)
            //{
            //    RTSConfiguration step = list2[i];

            //    DBM newDBM = step.DBM.Clone();
            //    newDBM.Conjunction(firstIdle);
            //    //foreach (ClockConstraints list in FirstProcess.IdleTimeConstraints)
            //    //{
            //    //    newDBM.AddConstraint(list.TimeID, list.Operator, list.Constant);
            //    //}
            //    step.DBM = newDBM;
                
            //    returnList.Add(step);
            //});)

            return list1;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList<string>(Variables, SecondProcess.GetGlobalVariables());

            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            return "(" + FirstProcess.ToString() + "|>" + SecondProcess.ToString() + ")";
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new Interrupt(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping));
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override void CalculateIdleTime(RTSConfiguration eStep, DBM delayDBM)
        //{
        //    FirstProcess.CalculateIdleTime(eStep, delayDBM);
        //    SecondProcess.CalculateIdleTime(eStep, delayDBM);
        //}

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    return Common.Classes.Ultility.Ultility.Union(FirstProcess.GetActiveTimer(), SecondProcess.GetActiveTimer());
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess first = FirstProcess.SetTimer(timerID, clocks);
            PARProcess second = SecondProcess.SetTimer(timerID, clocks);

            if (first == null && second == null)
            {
                return null;
            }
            
            if(first == null)
            {
                first = FirstProcess;
            }

            if(second == null)
            {
                second = SecondProcess;
            }
            
            return new Interrupt(first, second);
        }

        //public override bool IsTime()
        //{
        //    return FirstProcess.IsTime() || SecondProcess.IsTime();
        //}
    }
}