using System.Text;
using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARDeadlineProcess : PARProcess
    {
        public  PARProcess Process;
        public  Variable Parameter;
        public byte TimerID;

        public PARDeadlineProcess(PARProcess process, Variable parameter, byte timerID)
        {
            Process = process;
            Parameter = parameter;
            TimerID = timerID;         
            
            StringBuilder ID = new StringBuilder(Process.ProcessID);
            ID.Append(Constants.DEADLIME + "[" + Parameter + "]");

            ProcessID = PARDataStore.DataManager.InitializeProcessID(ID.ToString());
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = Process.GetGlobalVariables();
            //Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());

            return Variables;
        }

        public override List<string> GetChannels()
        {
            return Process.GetChannels();
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> steps = Process.MoveOneStep(eStep);

            // Get the parameter name
            string parameter_name = Parameter.ToString();
            // Add "clock <= param"

            // DEBUG ETIENNE
            eStep.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event != Constants.TERMINATION)
                {
                    vm.Process = new PARDeadlineProcess(vm.Process, Parameter, TimerID);
                }
            }

            return steps;
        }

    
        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            List<PARConfigurationWithChannelData> steps = Process.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                vm.Process = new PARDeadlineProcess(vm.Process, Parameter, TimerID);
            }

            return steps;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            List<PARConfiguration> steps = Process.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                vm.Process = new PARDeadlineProcess(vm.Process, Parameter, TimerID);
            }

            return steps;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + Process + " deadline[" + Parameter + "])";
            }
            else
            {
                return "(" + Process + " deadline[" + Parameter + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return Process.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARDeadlineProcess(Process.ClearConstant(constMapping), Parameter, TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return Process.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, Process.GetActiveTimer());//eStep));
        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = Process.SetTimer(timerID, clocks);

            // Get the parameter name
            string parameter_name = Parameter.ToString();

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddParInequality(TimerID, parameter_name);
                    return new PARDeadlineProcess(process, Parameter, TimerID);
                }

                clocks.AddParInequality(timerID, parameter_name);
                return new PARDeadlineProcess(process, Parameter, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddParInequality(TimerID, parameter_name);
                return null;
            }

            clocks.AddParInequality(timerID, parameter_name);
            return new PARDeadlineProcess(Process, Parameter, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return Process.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}