using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;

namespace PAT.PAR.LTS
{
    public abstract class PARProcess //: ProcessBase<RTSConfiguration>
    {
        public string ProcessID;

        /// <summary>
        /// returns all the possible parametric timed moves of the current process
        /// </summary>
        /// <returns></returns>
        public abstract List<PARConfiguration> MoveOneStep(PARConfiguration eStep);

        /// <summary>
        /// Get the set of global variables which may be accessed by this process. Notice that arrays will be flatened (for one level). 
        /// For instance, let leader[3] be an array, leader[0], leader[1] will be listed as two different variables. 
        /// </summary>
        /// <returns></returns>
        public virtual List<string> GetGlobalVariables()
        {
            return new List<string>(0);
        }

        /// <summary>
        /// Get the set of relevant channels.
        /// </summary>
        /// <returns></returns>
        public virtual List<string> GetChannels()
        {
            return new List<string>(0);
        }

        /// <summary>
        /// This method defines the logic for calculating the default alphabet of a process. That is, the set of events consitituting the 
        /// process expression with process reference unfolded ONCE! We found this to be intuitive at the moment.
        /// </summary>
        /// <returns></returns>
        public virtual HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return new HashSet<string>();
        }

        public virtual bool MustBeAbstracted()
        {
            return false;
        }

        /// <summary>
        /// clear global constants and process parameters; This method is the starting point of run-time execution of the process.
        /// </summary>
        /// <param name="constMapping"></param>
        /// <returns></returns>
        public abstract PARProcess ClearConstant(Dictionary<string, Expression> constMapping);

        ///// <summary>
        ///// returns all the possible synchronous output steps
        ///// </summary>
        ///// <returns></returns>
        public virtual List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            return new List<PARConfigurationWithChannelData>();
        }

        public virtual List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            return new List<PARConfiguration>(0);
        }

        public virtual PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            return null;
        }

        //public virtual List<byte> GetActiveTimer() //Configuration eStep)
        //{
        //    return new List<byte>(0);
        //}

        public virtual PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return null;
        }

        //public virtual bool IsTime()
        //{
        //    return false;
        //}
    }
}