﻿#define DEBUGETIENNE

using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.Ultility;

namespace PAT.PAR.LTS
{
    public class PARSimulationConfiguration : PARConfiguration
    {
        public Constraint SimConstraint;

        public PARSimulationConfiguration(PARProcess p, string e, string hiddenEvent, Valuation globalEnv, bool isDataOperation, bool isUrgent, Constraint constraint)
            : base(p, e, hiddenEvent, globalEnv, isDataOperation, isUrgent, constraint)
        {
        }

        public PARSimulationConfiguration(PARConfiguration config)
            : base(config.Process, config.Event, config.HidenEvent, config.GlobalEnv, config.IsDataOperation, config.IsUrgent, config.TheConstraint)
        {

        }

        public override ConfigurationBase[] MakeOneMove()
        {
            return MakeOneMoveLocal().ToArray();
        }



        public override void SetTimer()
        {
            SimConstraint = TheConstraint.Clone();

#if DEBUGETIENNE
            System.Console.Out.WriteLine("--------------------------------------------------");
            System.Console.Out.WriteLine("Entering PARSimulationConfiguration::SetTimer()");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

            NewClock = TheConstraint.GetNewClockID();

#if DEBUGETIENNE
            System.Console.Out.WriteLine("NewClock = " + Constraint.ClockToString(NewClock));
#endif

            List<byte> Clocks = new List<byte>();
            List<Constraint> Constraints = new List<Constraint>();

#if DEBUGETIENNE
            System.Console.Out.WriteLine("About to call Process.SetTimer()");
#endif
            PARProcess newProc = Process.SetTimer(NewClock, new ClockInfo(GlobalEnv, Clocks, Constraints));

#if DEBUGETIENNE
            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
#endif

#if DEBUGETIENNE
            System.Console.Out.WriteLine("About to keep clocks " + Constraint.SomeClockstoString(Clocks));
#endif
            TheConstraint = TheConstraint.KeepClocks(Clocks);

#if DEBUGETIENNE
            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
#endif

#if DEBUGETIENNE
            System.Console.Out.WriteLine("About to ensure that clocks " + Constraint.SomeClockstoString(Clocks) + " are nonnegative");
#endif
            // Necessary to add x >= 0 for all clocks in 'Clocks'!
            TheConstraint.EnsureClocksNonnegative(Clocks);

#if DEBUGETIENNE
            System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
#endif
            
            if (newProc != null)
            {
                Process = newProc;
            }

            // NEW TEST BY ETIENNE
            if (Clocks.Contains(NewClock) && !TheConstraint.ContainsClock(NewClock))
            {
#if DEBUGETIENNE
                System.Console.Out.WriteLine("TheConstraint does not contains " + Constraint.ClockToString(NewClock));
#endif
                TheConstraint.AddClock(NewClock);
#if DEBUGETIENNE
                System.Console.Out.WriteLine("TheConstraint = " + TheConstraint.ToString());
#endif
            }
            else
            {
#if DEBUGETIENNE
                if (Clocks.Contains(NewClock))
                {
                    System.Console.Out.WriteLine("TheConstraint contains " + Constraint.ClockToString(NewClock) + " already");
                }
                else
                {
                    System.Console.Out.WriteLine("NewClock " + Constraint.ClockToString(NewClock) + " has not been used: forget it.");
                }  
#endif

            }

            // Delay the constraint
            TheConstraint.Delay();

#if DEBUGETIENNE
            System.Console.Out.WriteLine("Delayed TheConstraint : " + TheConstraint.ToString());
#endif

            // Add inequalities
            TheConstraint.AddConstraints(Constraints);
#if DEBUGETIENNE
            System.Console.Out.WriteLine("Added inequalities to TheConstraint : " + TheConstraint.ToString());
#endif

#if DEBUGETIENNE
            System.Console.Out.WriteLine("Leaving PARSimulationConfiguration::SetTimer()");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            System.Console.Out.WriteLine("--------------------------------------------------");
#endif
        }




        public override List<PARConfiguration> MakeOneMoveLocal()
        {
#if DEBUGETIENNE
            System.Console.Out.WriteLine("****************************************************************************************************");
            System.Console.Out.WriteLine("Entering PARSimulationConfiguration::MakeOneMoveLocal()");
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
            if (this != null)
            {
                System.Console.Out.WriteLine(this.ToString());
            }
            System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

            List<PARConfiguration> list = Process.MoveOneStep(this);

#if DEBUGETIENNE
            System.Console.Out.WriteLine("Perormed Process.MoveOneStep(this)");
#endif

            List<PARConfiguration> returnListAtomic = new List<PARConfiguration>();
            List<PARConfiguration> returnListNonAtomic = new List<PARConfiguration>();

            bool hasUrgent = false;

            foreach (PARConfiguration configuration in list)
            {

#if DEBUGETIENNE
                System.Console.Out.WriteLine("--Considering a configuration");
                System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                System.Console.Out.WriteLine(configuration.ToString());
                System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

                // Separation between atomic and non-atomic configurations
                if (configuration.IsAtomic)
                {
#if DEBUGETIENNE
                    System.Console.Out.WriteLine("  This is an atomic configuration");
#endif

                    configuration.TheConstraint = configuration.TheConstraint.Clone();
//#if DEBUGETIENNE
//                    System.Console.Out.WriteLine("  About to add inequality '"
//                        + Constraint.ClockToString(NewClock) + "=0'");
//#endif

                    // ETIENNE SAYS: WHY?!
                    // configuration.TheConstraint.AddInequality(NewClock, TimerOperationType.Equals, 0);
                    // configuration.TheConstraint.AddClock(NewClock);

                    if (configuration.TheConstraint.IsSatisfiable())
                    {
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  Constraint is satisfiable");
#endif
                        PARSimulationConfiguration newConfig = new PARSimulationConfiguration(configuration);
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  About to call SetTimer()");
#endif
                        newConfig.SetTimer();
                        returnListAtomic.Add(newConfig);
                    }
                }
                else
                {
#if DEBUGETIENNE
                    System.Console.Out.WriteLine("  This is an non-atomic configuration");
#endif
                    if (configuration.TheConstraint.IsSatisfiable())
                    {
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  Constraint is satisfiable");
#endif
                        if (configuration.IsUrgent)
                        {
#if DEBUGETIENNE
                            System.Console.Out.WriteLine("  Configuration is urgent");
#endif
                            hasUrgent = true;
                        }

                        returnListNonAtomic.Add(configuration);
                    }
                }
            }

            // if atomic, then only return atomic
            if (returnListAtomic.Count > 0)
            {
#if DEBUGETIENNE
                System.Console.Out.WriteLine("Found atomic configurations: return atomic");
#endif
                return returnListAtomic;
            }

            else
            {
                // THIS DISTINCTION SEEMS UNNECESSARY!
                if (hasUrgent)
                {
#if DEBUGETIENNE
                    System.Console.Out.WriteLine("  Found urgent configurations");
#endif
                    List<PARConfiguration> toReturn = new List<PARConfiguration>();

                    foreach (PARConfiguration configuration in returnListNonAtomic)
                    {
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("--Considering a urgent configuration");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(configuration.ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif
                        // NECESSARY ?
                        //PARConfiguration.TheConstraint = configuration.TheConstraint.Clone();
                        // IMPOSSIBLE!
                        //configuration.TheConstraint.AddInequality(NewClock, TimerOperationType.LessThanOrEqualTo, timelimit);

                        // TEST BECAME UNNECESSARY
                        //if (configuration.TheConstraint.IsSatisfiable())
                        //{
                        PARSimulationConfiguration newConfig = new PARSimulationConfiguration(configuration);
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  About to set timer");
#endif
                        newConfig.SetTimer();
                        toReturn.Add(newConfig);
                        //}
                    }

                    IsDeadLock = toReturn.Count == 0;
                    return toReturn;
                }
                else
                {
#if DEBUGETIENNE
                    System.Console.Out.WriteLine("No urgent configuration found.");
#endif
                    //foreach (PARConfiguration configuration in returnListNonAtomic)
                    //{
                    //    configuration.SetTimer();
                    //}
                    for (int i = 0; i < returnListNonAtomic.Count; i++)
                    {
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("--Considering a non-urgent configuration");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(returnListNonAtomic[i].ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif
                        PARSimulationConfiguration newConfig = new PARSimulationConfiguration(returnListNonAtomic[i]);
#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  About to call SetTimer()");
#endif
                        newConfig.SetTimer();

#if DEBUGETIENNE
                        System.Console.Out.WriteLine("  After SetTimer, configuration becomes: ");
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                        System.Console.Out.WriteLine(newConfig.ToString());
                        System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
#endif

                        returnListNonAtomic[i] = newConfig;
                    }

                    IsDeadLock = returnListNonAtomic.Count == 0;
#if DEBUGETIENNE
                    System.Console.Out.WriteLine("Leaving PARSimulationConfiguration::MakeOneMoveLocal()");
                    System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                    if (this != null)
                    {
                        System.Console.Out.WriteLine(this.ToString());
                    }
                    System.Console.Out.WriteLine(". . . . . . . . . . . . . . . . . . . . ");
                    System.Console.Out.WriteLine("****************************************************************************************************");
#endif
                    return returnListNonAtomic;
                }
            }
        }




        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine("The process is:");
            sb.AppendLine(Process.ToString());


            if (!GlobalEnv.IsEmpty())
            {
                sb.AppendLine();
                sb.AppendLine("The environment is:");
                sb.AppendLine(GlobalEnv.ToString());

            }

            //List<byte> timers = Process.GetActiveTimer();

            //sb.AppendLine();
            //sb.AppendLine("The active clocks SIM are:");
            //sb.AppendLine(SimConstraint.ClocksToString());

            ////sb.AppendLine();
            ////sb.AppendLine("The active clocks are (according to ClockInfo):");
            ////sb.AppendLine(Constraint.SomeClockstoString(Clocks));

            //sb.AppendLine();
            //sb.AppendLine("The constraint SIM is:");
            //sb.AppendLine(SimConstraint.ToString());

            //sb.AppendLine();
            //sb.AppendLine("The constraint SIM onto the parameters is:");
            //sb.AppendLine(SimConstraint.HideAllClocks().ToString()); // Clone very important here!!

            sb.AppendLine();
            sb.AppendLine("The active clocks are:");
            sb.AppendLine(TheConstraint.ClocksToString());

            string theParameters = "";
            foreach (string name in Specification.Parameters)
            {
                theParameters += name + " - ";
            }
            sb.AppendLine();
            sb.AppendLine("The parameters are: " + theParameters);


            sb.AppendLine();
            sb.AppendLine("The constraint is:");
            sb.AppendLine(TheConstraint.ToString());

            sb.AppendLine();
            sb.AppendLine("The constraint onto the parameters is:");
            sb.AppendLine(TheConstraint.HideAllClocks().ToString()); // Clone very important here!!

            return sb.ToString();
        }
    }
}
