﻿using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;

namespace PAT.PAR.LTS
{
    class PPL
    {
        [DllImport("kernel32", CharSet = CharSet.Ansi, ExactSpelling = true, SetLastError = true)]
        static extern IntPtr GetProcAddress(IntPtr hModule, string procName);

        [DllImport("kernel32", SetLastError = true)]
        static extern IntPtr LoadLibrary(string lpFileName);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern bool FreeLibrary([In] IntPtr hModule);

        public delegate int PPL_isSatisfiable(string polyhedron);
        public delegate int PPL_comparison(string polyhedron1, string polyhedron2);
        public delegate IntPtr PPL_simplify(string polyhedron);
        public delegate IntPtr PPL_2to1(string string1, string string2);

        // Converts an int to a bool (because bool from C++ are not passed properly to C#)
        private static bool bool_of_int(int i){
            if (i == 1)
            {
                return true;
            }
            else if (i == 0)
            {
                return false;
            }
            throw new Exception("Error when converting an int to a bool. Expected: 0 or 1." 
                + " Passed: " + i
            );
        }

        public static bool isSatisfiable(string polyhedron)
        {
            // Open the library
            IntPtr ppl_pointer = LoadLibrary("ppl.dll");

            // Call the function
            IntPtr function_pointer = GetProcAddress(ppl_pointer, "isSatisfiable");
            PPL_isSatisfiable isSatisfiable = (PPL_isSatisfiable)Marshal.GetDelegateForFunctionPointer(function_pointer, typeof(PPL_isSatisfiable));
            bool result = bool_of_int(isSatisfiable(polyhedron));

            // Close the library
            FreeLibrary(ppl_pointer);

            return result;
        }


        private static string op2to1(string string1, string string2, string operation)
        {
            // Open the library
            IntPtr ppl_pointer = LoadLibrary("ppl.dll");

            // Call the function
            IntPtr function_pointer = GetProcAddress(ppl_pointer, operation);
            PPL_2to1 op = (PPL_2to1)Marshal.GetDelegateForFunctionPointer(function_pointer, typeof(PPL_2to1));
            string result = Marshal.PtrToStringAnsi(op(string1, string2));

            // Close the library
            FreeLibrary(ppl_pointer);

            return result;
        }

        public static string intersection(string polyhedron1, string polyhedron2)
        {
            return op2to1(polyhedron1, polyhedron2, "intersection");
        }

        public static string hide(string polyhedron, string variables)
        {
            return op2to1(polyhedron, variables, "hide");
        }

        public static string time_elapsing(string polyhedron, string variables)
        {
            return op2to1(polyhedron, variables, "time_elapsing");
        }
        

        
        private static bool comp(string polyhedron1, string polyhedron2, string operation)
        {
            // Open the library
            IntPtr ppl_pointer = LoadLibrary("ppl.dll");

            // Call the function
            IntPtr function_pointer = GetProcAddress(ppl_pointer, operation);
            PPL_comparison comp = (PPL_comparison)Marshal.GetDelegateForFunctionPointer(function_pointer, typeof(PPL_comparison));
            bool result = bool_of_int(comp(polyhedron1, polyhedron2));

            // Close the library
            FreeLibrary(ppl_pointer);

            return result;
        }


        

        public static bool equals(string polyhedron1, string polyhedron2)
        {
            return comp(polyhedron1, polyhedron2, "equals");
        }

        public static bool contains(string polyhedron1, string polyhedron2)
        {
            return comp(polyhedron1, polyhedron2, "contains");
        }


        public static string simplify(string polyhedron)
        {
            // Open the library
            IntPtr ppl_pointer = LoadLibrary("ppl.dll");

            // Call the function
            IntPtr function_pointer = GetProcAddress(ppl_pointer, "simplify");
            PPL_simplify simplify = (PPL_simplify)Marshal.GetDelegateForFunctionPointer(function_pointer, typeof(PPL_simplify));
            string result = Marshal.PtrToStringAnsi(simplify(polyhedron));

            // Close the library
            FreeLibrary(ppl_pointer);

            return result;
        }
    }
}
