﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;

namespace PAT.PAR.LTS
{
    public class Polyhedron
    {

        // Interface functions between bool and integers
        // Important because "sometimes" booleans are not understood the same between C# and C++
        private int int_of_bool(bool b){
            if (b)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }

        private bool bool_of_int(int i){
            if (i == 1)
            {
                return true;
            }
            else if (i == 0)
            {
                return false;
            }
            else
            {
                throw new Exception("Illegal integer found; only 0 and 1 are expected.");
            }
        }


        #region dllimportfun


        [DllImport("polyhedra.dll")]
        static private extern void SeedRand(int seed);

        [DllImport("polyhedra.dll")]
        static private extern IntPtr TrueConstraint();

        [DllImport("polyhedra.dll")]
        static private extern IntPtr FalseConstraint();

        [DllImport("polyhedra.dll")]
        static private extern void Free(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern IntPtr Clone(IntPtr obj);


        [DllImport("polyhedra.dll")]
        static private extern string GetId(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern int IsSatisfiable(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern int IsTrue(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern int IsFalse(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern int GetHighestDimension(IntPtr obj);

        [DllImport("polyhedra.dll")]
        static private extern int Incompatible(IntPtr obj, int random_selection, int[] pi0, int len, IntPtr negated_inequality);


        [DllImport("polyhedra.dll")]
        static private extern void Add_L_Const(IntPtr obj, int variable, int constant);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Leq_Const(IntPtr obj, int variable, int constant);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Eq_Const(IntPtr obj, int variable, int constant);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Geq_Const(IntPtr obj, int variable, int constant);
        [DllImport("polyhedra.dll")]
        static private extern void Add_G_Const(IntPtr obj, int variable, int constant);

        [DllImport("polyhedra.dll")]
        static private extern void Add_L_Var(IntPtr obj, int variable1, int variable2);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Leq_Var(IntPtr obj, int variable1, int variable2);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Eq_Var(IntPtr obj, int variable1, int variable2);
        [DllImport("polyhedra.dll")]
        static private extern void Add_Geq_Var(IntPtr obj, int variable1, int variable2);
        [DllImport("polyhedra.dll")]
        static private extern void Add_G_Var(IntPtr obj, int variable1, int variable2);

        [DllImport("polyhedra.dll")]
        static private extern void Intersection(IntPtr obj, IntPtr c2);

        [DllImport("polyhedra.dll")]
        static private extern void Hide(IntPtr obj, int[] var_set, int len);

        [DllImport("polyhedra.dll")]
        static private extern void TimeElapsing(IntPtr obj, int[] var_set, int len);

        [DllImport("polyhedra.dll")]
        static private extern void Rename(IntPtr obj, int offset, int[] var_set, int len);

        #endregion

        private IntPtr poly = IntPtr.Zero;
        private bool disposed = false;

        public Polyhedron()
        {
            poly = TrueConstraint();
        }

        public Polyhedron FalsePolyhedron()
        {
            Polyhedron p = new Polyhedron();
            p.poly = FalseConstraint();
            return p;
        }

        ~Polyhedron()
        {
            Dispose(false);
        }
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if ( !this.disposed )
            {
                Free(poly);
                poly = IntPtr.Zero;

            }
            disposed = true;
        }

        public Polyhedron Clone()
        {
            Polyhedron cloned = new Polyhedron();
            cloned.poly = Clone(poly);
            return cloned;
        }

        public string GetId()
        {
            //IntPtr polyCopy = Clone(poly);
            //string result = GetId(polyCopy);
            //Free(polyCopy);
            //return result;
            return GetId(poly);
        }

        public bool IsSatisfiable()
        {
            //IntPtr polyCopy = Clone(poly);
            //System.Console.Out.WriteLine("IsSat ---> ID = " + GetId());
            //bool result = IsSatisfiable(polyCopy);
            //System.Console.Out.WriteLine("IsSat ---> result = " + result);
            ////bool result2 = IsSatisfiable(poly);
            ////if (result != result2)
            ////{
            ////    throw new Exception("Bug in sat test! Sat(copy) = " + result + " ; Sat(this) = " + result2 + " .");
            ////}
            //Free(polyCopy);
            //return result;
            return bool_of_int(IsSatisfiable(poly));
        }

        public bool IsTrue()
        {
            //IntPtr polyCopy = Clone(poly);
            //bool result = IsTrue(polyCopy);
            //Free(polyCopy);
            //return result;
            return bool_of_int(IsTrue(poly));
        }

        public bool IsFalse()
        {
            //IntPtr polyCopy = Clone(poly);
            //bool result = IsFalse(polyCopy);
            //Free(polyCopy);
            //return result;
            return bool_of_int(IsFalse(poly));
        }

        public int GreatestVariable()
        {
            //IntPtr polyCopy = Clone(poly);
            //int result = GetHighestDimension(polyCopy);
            //Free(polyCopy);
            //return result;
            return GetHighestDimension(poly);
        }

        /**
         * Test the compatibility w.r.t. a reference valuation
         * Reference valuation 'pi0' is built as follows:
         * - variable i has reference valuation pi0[i], for 0 <= i <= pi0_len-1 
         * If incompatible: select randomly (if 'random_selection' is true), or stop as soon as one is found (otherwise)
         * Assigns the negation of an incompatible inequality to 'negated_inequality', or 'false' if p is compatible
         * The returned bool indicates whether the choice was random:
         * - it is true if 'random_selection' is true and several incompatible inequalities were found, and false otherwise
         * C++ layer will (try to) raise an exception if the constraint is unsatisfiable
         * C++ layer will (try to) raise an exception if a variable used in the constraint does not appear in pi0 (i.e., if variable >= pi0_len)
         */
        public bool Incompatible(bool random_selection, int[] pi0, Polyhedron negated_inequality)
        {
            //IntPtr polyCopy = Clone(poly);
            //bool result = Incompatible(polyCopy, random_selection, pi0, pi0.Length, negated_inequality.poly);
            //Free(polyCopy);
            //return result;
            return bool_of_int(Incompatible(poly, int_of_bool(random_selection), pi0, pi0.Length, negated_inequality.poly));
        }


        public void AddLConst(byte clock, int constant)
        {
            Add_L_Const(poly, clock, constant);
        }
        public void AddLeqConst(byte clock, int constant)
        {
            Add_Leq_Const(poly, clock, constant);
        }
        public void AddEqConst(byte clock, int constant)
        {
            Add_Eq_Const(poly, clock, constant);
        }
        public void AddGeqConst(byte clock, int constant)
        {
            Add_Geq_Const(poly, clock, constant);
        }
        public void AddGConst(byte clock, int constant)
        {
            Add_G_Const(poly, clock, constant);
        }

        public void AddLVar(byte var1, byte var2)
        {
            Add_L_Var(poly, var1, var2);
        }
        public void AddLeqVar(byte var1, byte var2)
        {
            Add_Leq_Var(poly, var1, var2);
        }
        public void AddEqVar(byte var1, byte var2)
        {
            Add_Eq_Var(poly, var1, var2);
        }
        public void AddGeqVar(byte var1, byte var2)
        {
            Add_Geq_Var(poly, var1, var2);
        }
        public void AddGVar(byte var1, byte var2)
        {
            Add_G_Var(poly, var1, var2);
        }

        public void Intersection(Polyhedron p2)
        {
            Intersection(poly, p2.poly);
        }

        public void Hide(int[] var_set)
        {
            Hide(poly, var_set, var_set.Length);
        }

        public void TimeElapsing(int[] var_set)
        {
            TimeElapsing(poly, var_set, var_set.Length);
        }

        // renaming is an array index -> value, such that var 'index + offset' will be renamed by 'value'
        // renaming must be a BIJECTION; if not, the behavior is completely unspecified
        public void Rename(int offset, int[] renaming)
        {
            Rename(poly, offset, renaming, renaming.Length);
        }

        public static void Seed()
        {
            Random rndNumbers = new Random();
            int rndNumber = rndNumbers.Next();
            SeedRand(rndNumber);
        }
    }
}

