using System;
using System.Collections.Generic;
using System.Text;
using Antlr.Runtime;
using Antlr.Runtime.Tree;
using PAT.Common;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using PAT.Common.Classes.ModuleInterface;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;
using PAT.PAR.Ultility;

namespace PAT.PAR.LTS
{
    /// <summary>
    /// The specification class a collection of the definitions, properties and alphasets
    /// Each of the user's input correspons to a user's input file
    /// </summary>
    public class Specification : SpecificationBase
    {
        //entry:(name::string, defintion::CSPProcess)
        public Dictionary<string, Definition> DefinitionDatabase = new Dictionary<string, Definition>(16);
        public Dictionary<string, ChannelQueue> ChannelDatabase = new Dictionary<string, ChannelQueue>(8);

        public static HashSet<string> Parameters = new HashSet<string>();

        public Valuation SpecValuation = new Valuation();

        //todo add it to shared object.
        //public static bool TimedRefinementAssertion = false;
        //public static int TimedRefinementClockCeiling;
        //public static int TimedRefinementClockFloor;

        public SharedDataObjects SharedData;

        // Init constraint on the parameters, always valid (i.e., all states will contain this constraint)
        public Expression ParameterConstraint;

        //constructor used for the console, testing for sequencial access.
        public Specification(string spec) : base(spec, null)
        {
            
            SharedData = new SharedDataObjects();

            LockSpecificationData();

            ParseSpec(spec, "");
        }


        public Specification(string spec, string option, string filePath) : base(spec, filePath)
        {
            SharedData = new SharedDataObjects();
            Common.Classes.Ultility.Ultility.LockSharedData(this);

            try
            {
                ParseSpec(spec, option);

                SharedData.HasSyncrhonousChannel = HasSyncrhonousChannel;
                SharedData.SyncrhonousChannelNames = SyncrhonousChannelNames;
                SharedData.HasAtomicEvent = HasAtomicEvent;
                SharedData.HasCSharpCode = HasCSharpCode;
                SharedData.LocalVars = Valuation.HiddenVars;
                SharedData.Parameters = new HashSet<string>(Parameters);

            }
            catch (Exception ex)
            {
                throw ex;
            }
            finally
            {
                Common.Classes.Ultility.Ultility.UnLockSharedData(this);
            }
        }

        protected virtual void ParseSpec(string spec, string option)
        {
            IsParsing = true;

            //parse the specification
            //build the grammar tree
            ICharStream input = new ANTLRStringStream(spec);
            PARTreeLexer lex = new PARTreeLexer(input);
            CommonTokenStream tokens = new CommonTokenStream(lex);
            PARTreeParser parser = new PARTreeParser(tokens);
            parser.Spec = this;
            PARTreeParser.specification_return r = parser.specification();


            //====================START================================
            //add the included file into the model and parse again
            int count = 0;

            while (IncludeFiles.Count > count)
            {
                count = IncludeFiles.Count;

                string completeSpec = GetCompleteSpecFromIncludeFiles();              

                input = new ANTLRStringStream(completeSpec);
                lex = new PARTreeLexer(input);
                tokens = new CommonTokenStream(lex);
                parser = new PARTreeParser(tokens);
                parser.Spec = this;
                r = parser.specification();
            }
            //====================END================================        

            if (parser.HiddenVars.Count > 0)
            {
                Valuation.HiddenVars = parser.HiddenVars;
            }

            Common.Ultility.ParsingUltility.CheckIsParsingComplete(tokens, r.Tree);

            CommonTreeNodeStream nodes = new CommonTreeNodeStream(r.Tree);
            nodes.TokenStream = tokens;

            //construct the internal representation of the tree via a tree walker
            PARTreeWalker walker = new PARTreeWalker(nodes);
            walker.GlobalVarNames = parser.GlobalVarNames;
            walker.GlobalParameterNames = parser.GlobalParameterNames;    
            walker.GlobalConstNames = parser.GlobalConstNames;
            walker.GlobalRecordNames = parser.GlobalRecordNames;
            walker.ChannelNames = parser.ChannelNames;
            walker.LTLStatePropertyNames = parser.LTLStatePropertyNames;
            walker.DefinitionNames = parser.DefinitionNames;
            walker.IsParameterized = parser.IsParameterized;
            walker.HasArbitraryProcess = parser.HasArbitraryProcess;

            walker.Spec = this;
            walker.program(option);

            IsParsing = false;

            GlobalConstantDatabase = walker.ConstantDatabase;

            StaticAnalysis();

            foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
            {
                List<string> gVar = pair.Value.GlobalVars;

                int i = 0;
                while (i < gVar.Count)
                {
                    if (SpecValuation.Variables != null && !SpecValuation.Variables.ContainsKey(gVar[i]))
                    {
                        gVar.RemoveAt(i);
                    }
                    else
                    {
                        i++;
                    }
                }
            }

            //get the relevant channels; 
            if (ChannelDatabase.Count > 0)
            {
                SyncrhonousChannelNames = new List<string>(0);
                Dictionary<string, ChannelQueue> newChannelDatabase = new Dictionary<string, ChannelQueue>();

                foreach (KeyValuePair<string, ChannelQueue> pair in ChannelDatabase)
                {
                    if (pair.Value.Size == 0)
                    {
                        SyncrhonousChannelNames.Add(pair.Key);
                    }
                    else
                    {
                        newChannelDatabase.Add(pair.Key, pair.Value);
                    }
                }

                SpecValuation.Channels = newChannelDatabase;
                HasSyncrhonousChannel = SyncrhonousChannelNames.Count > 0;
            }

            // VERY IMPORTANT (must be done at least one, and prefarbly only once, before starting analysis)
            Constraint.InitParameters();

            foreach (KeyValuePair<string, AssertionBase> entry in AssertionDatabase)
            {
                entry.Value.Initialize(this);
            }

            CheckVariableRange();
        }

        protected void CheckVariableRange()
        {
            foreach (KeyValuePair<string, Declaration> declaration in DeclaritionTable)
            {
                if (declaration.Value.DeclarationType == DeclarationType.Variable)
                {
                    Valuation.CheckVariableRange(declaration.Key, SpecValuation.Variables[declaration.Key], declaration.Value.DeclarationToken.Line, declaration.Value.DeclarationToken.CharPositionInLine);
                }
            }
        }
        /// <summary>
        /// Perform static analysis to identity things like
        /// Whether a definition must be abstraction
        /// what are the relevant variables
        /// what are the relevant channels
        /// </summary>
        private void StaticAnalysis()
        {
            //Create the definition graph by using the names.
            foreach (Definition definition in DefinitionDatabase.Values)
            {
                foreach (string name in definition.SubDefinitionNames)
                {
                    if (DefinitionDatabase.ContainsKey(name))
                    {
                        definition.SubDefinitions.Add(DefinitionDatabase[name]);
                    }
                }
                definition.SubDefinitionNames = null;
            }


            foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
            {
                pair.Value.StaticAnalysis();
            }

            //update MustAbstract
            bool modified = true;
            while (modified)
            {
                modified = false;
                foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
                {
                    if (!pair.Value.MustAbstract)
                    {
                        List<Definition> calls = pair.Value.SubDefinitions;

                        foreach (Definition s in calls)
                        {
                            if (s.MustAbstract)
                            {
                                pair.Value.MustAbstract = true;
                                modified = true;
                                break;
                            }
                        }
                    }
                }
            }

            ////update MustAbstract
            //modified = true;
            //while (modified)
            //{
            //    modified = false;
            //    foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
            //    {
            //        //if (!pair.Value.IsTimed)
            //        //{
            //            List<Definition> calls = pair.Value.SubDefinitions;

            //            foreach (Definition s in calls)
            //            {
            //                //if (s.IsTimed)
            //                //{
            //                    pair.Value.IsTimed = true;
            //                    modified = true;
            //                    break;
            //                //}
            //            }
            //        //}
            //    }
            //}


            modified = true;
            while (modified)
            {
                modified = false;
                foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
                {
                    if (pair.Value.AlphabetsCalculable)
                    {
                        List<Definition> calls = pair.Value.SubDefinitions;

                        foreach (Definition s in calls)
                        {
                            if (!s.AlphabetsCalculable)
                            {
                                pair.Value.AlphabetsCalculable = false;
                                modified = true;
                                break;
                            }
                        }
                    }
                }
            }

            modified = true;
            while (modified)
            {
                modified = false;
                foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
                {
                    if (pair.Value.AlphabetsCalculable && pair.Value.AlphabetEvents == null)
                    {
                        List<Definition> calls = pair.Value.SubDefinitions;

                        foreach (Definition s in calls)
                        {
                            foreach (string var in s.Alphabets)
                            {
                                if (!pair.Value.Alphabets.Contains(var))
                                {
                                    pair.Value.Alphabets.Add(var);
                                    modified = true;
                                }
                            }
                        }
                    }
                }
            }


            modified = true;
            while (modified)
            {
                modified = false;
                foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
                {

                    List<Definition> calls = pair.Value.SubDefinitions;

                    foreach (Definition s in calls)
                    {
                        foreach (string channel in s.Channels)
                        {
                            if (!pair.Value.Channels.Contains(channel))
                            {
                                pair.Value.Channels.Add(channel);
                                modified = true;
                            }
                        }
                    }
                }
            }

            modified = true;
            while (modified)
            {
                modified = false;
                foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
                {

                    List<Definition> calls = pair.Value.SubDefinitions;

                    foreach (Definition s in calls)
                    {
                        foreach (string var in s.GlobalVars)
                        {
                            if (!pair.Value.GlobalVars.Contains(var))
                            {
                                pair.Value.GlobalVars.Add(var);
                                modified = true;
                            }
                        }
                    }
                }
            }

            return;
        }

        public override ConfigurationBase SimulationInitialization(string startingProcess)
        {
            Definition def = DefinitionDatabase[startingProcess];
            DefinitionRef defref = new DefinitionRef(def.Name, new Expression[0]) { Def = def };
            
            // ETIENNE SAYS: no need in PAR
            // PARSimulationConfiguration step = new PARSimulationConfiguration(defref, Constants.INITIAL_EVENT, null, SpecValuation, false, false, Constraint.constraintWithNonNegativeParameters());  // trueConstraint
            PARConfiguration step = new PARConfiguration(defref, Constants.INITIAL_EVENT, null, SpecValuation, false, false, Constraint.constraintWithNonNegativeParameters());  // trueConstraint
            
            step.SetTimer();
            return step;
        }

        public override string GetSpecification()
        {
            StringBuilder sb = new StringBuilder();

            if (GlobalConstantDatabase != null && GlobalConstantDatabase.Count > 0)
            {
                sb.AppendLine("//====================Global Constants====================");
                foreach (KeyValuePair<string, Expression> pair in GlobalConstantDatabase)
                {
                    if (pair.Value is ExpressionValue)
                    {
                        sb.AppendLine("#define " + pair.Key + " " + pair.Value.ExpressionID + ";");
                    }
                    else
                    {
                        sb.AppendLine("#define " + pair.Key + " " + pair.Value.ToString() + ";");
                    } 
                }
            }

            if (SpecValuation.Variables != null)
            {
                sb.AppendLine("//====================Global Variable Definitions====================");
                foreach (StringDictionaryEntryWithKey<ExpressionValue> pair in SpecValuation.Variables._entries)
                {
                    if (pair != null)
                    {

                        string range = "";
                        if (Valuation.VariableLowerBound.ContainsKey(pair.Key))
                        {
                            range = ":{" + Valuation.VariableLowerBound.GetContainsKey(pair.Key) + "..";
                        }

                        if (Valuation.VariableUpperLowerBound.ContainsKey(pair.Key))
                        {
                            if (range == "")
                            {
                                range = ":{.." + Valuation.VariableUpperLowerBound.GetContainsKey(pair.Key) + "}";
                            }
                            else
                            {
                                range += Valuation.VariableUpperLowerBound.GetContainsKey(pair.Key) + "}";
                            }
                        }
                        else
                        {
                            if (range != "")
                            {
                                range += "}";
                            }
                        }

                        if (Valuation.HiddenVars != null && Valuation.HiddenVars.ContainsKey(pair.Key))
                        {
                            sb.AppendLine("hvar " + pair.Key + range + " = " + pair.Value + ";");
                        }
                        else
                        {
                            sb.AppendLine("var " + pair.Key + range + " = " + pair.Value + ";");
                        }

                    }
                }
            }



            if (ChannelDatabase.Count > 0)
            {
                sb.AppendLine("//====================Channel Definitions====================");
                foreach (KeyValuePair<string, ChannelQueue> pair in ChannelDatabase)
                {
                    sb.AppendLine("channel " + pair.Key + " " + pair.Value.Size + ";");
                }
            }


            if (Parameters.Count > 0)
            {
                sb.AppendLine("//====================Parameter Definitions====================");
                foreach (string par in Parameters)
                {
                    sb.AppendLine("parameter " + par + ";");
                }
            }

            sb.AppendLine("//====================Process Definitions====================");
            foreach (Definition entry in DefinitionDatabase.Values)
            {
                sb.AppendLine(entry.GetFullDefinition());
            }


            sb.AppendLine("\r\n");
            //sb.AppendLine("\r\n//============================Alphabets Definitions============================");
            foreach (Definition entry in DefinitionDatabase.Values)
            {
                if (entry.AlphabetEvents != null)
                {
                    sb.AppendLine("#alphabet " + entry.Name + " {" + Common.Classes.Ultility.Ultility.PPStringList(entry.AlphabetEvents) + "};");
                }
            }

            if(ParameterConstraint != null)
            {
                sb.AppendLine("constraint " + ParameterConstraint + ";");
            }

            if (DeclarationDatabase.Count > 0)
            {
                sb.AppendLine("\r\n//============================Declarations============================");

                foreach (KeyValuePair<string, Expression> entry in DeclarationDatabase)
                {
                    sb.AppendLine("#define " + entry.Key + " " + entry.Value + ";");
                }
            }

            if (AssertionDatabase.Count > 0)
            {
                sb.AppendLine("\r\n//============================Asserstion Definitions============================");

                foreach (KeyValuePair<string, AssertionBase> entry in AssertionDatabase)
                {
                    if (entry.Value is PARInverseMethod || entry.Value is PARSynthesisReachability || entry.Value is PARSynthesisRefinement)
                    {
                        sb.AppendLine("#synthesize " + entry.Key + ";");
                    }
                    else
                    {
                        sb.AppendLine("#assert " + entry.Key + ";");    
                    }
                    
                }
            }

            return sb.ToString();
        }

        public void ClearDatabase()
        {
            //HasFairEvent = false;
            DeclarationDatabase.Clear();
            DefinitionDatabase.Clear();
            AssertionDatabase.Clear();
            ChannelDatabase.Clear();
        }

        public override List<string> GetProcessNames()
        {
            List<string> sb = new List<string>(DefinitionDatabase.Count);

            foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
            {
                if (pair.Value.LocalVariables.Length == 0)
                {
                    sb.Add(pair.Key);
                }
            }

            sb.Sort();
            return sb;
        }

        public override List<string> GetAllProcessNames()
        {
            List<string> sb = new List<string>(DefinitionDatabase.Count);

            foreach (KeyValuePair<string, Definition> pair in DefinitionDatabase)
            {
                sb.Add(pair.Key);
            }

            sb.Sort();
            return sb;
        }

        public override List<string> GetGlobalVarNames()
        {
            List<string> sb = new List<string>();
            
            if (SpecValuation.Variables != null)
            {
                foreach (StringDictionaryEntryWithKey<ExpressionValue> pair in SpecValuation.Variables._entries)
                {
                    if (pair != null)
                    {
                        sb.Add(pair.Key);
                    }
                    
                }
            }

            sb.Sort();
            return sb;
        }

        public override List<string> GetChannelNames()
        {
            List<string> sb = new List<string>(this.ChannelDatabase.Count);

            foreach (KeyValuePair<string, ChannelQueue> pair in ChannelDatabase)
            {
                sb.Add(pair.Key);
            }

            sb.Sort();
            return sb;
        }

        public override string[] GetParameterNames(string process)
        {
            if (DefinitionDatabase.ContainsKey(process))
            {
                return DefinitionDatabase[process].LocalVariables;
            }
            else
            {
                return null;
            }
        }

        public PARProcess GetProcess(string name)
        {
            if (DefinitionDatabase.ContainsKey(name))
            {
                return DefinitionDatabase[name].Process;
            }
            else
            {
                return null;
            }
        }


        //Lock the specification data, this method is called before the simulation and verification
        public override void LockSpecificationData()
        {

            PARDataStore.DataManager = this.SharedData.DataManager;

            SharedData.LockSpecificationData();

            Specification.Parameters = SharedData.Parameters;

            // Constraint.NonNegativeParametersConstraint = SharedData.NonNegativeParametersConstraint;
            // Constraint.InitParameters();
            ////always false
            ////AssertionBase.ApplyNormalizationInTimedRefinementAssertion = false;
            ////Specification.TimedRefinementAssertion = this.SharedData.TimedRefinementAssertion;
            ////Specification.TimedRefinementClockCeiling = this.SharedData.TimedRefinementClockCeiling;
            ////Specification.TimedRefinementClockFloor = this.SharedData.TimedRefinementClockFloor;


        }

        //UnLock the specification data, this method is called after the simulation and verification
        //This is a special case for PAR module, because most of the variables' values are decided after parsing and the values are not changed
        //during the execution of the model. However, This is not true for NonNegativeParametersConstraint.
        //So I created a method for this parameter.
        public override void UnLockSpecificationData()
        {
            // SharedData.NonNegativeParametersConstraint = Constraint.NonNegativeParametersConstraint;
        }
    }


    public sealed class SharedDataObjects : SharedDataObjectBase
    {

        public PARDataStore DataManager;

        public bool TimedRefinementAssertion;
        public int TimedRefinementClockCeiling;
        public int TimedRefinementClockFloor;
        public HashSet<string> Parameters;

        public SharedDataObjects()
        {
            DataManager = new PARDataStore();

            TimedRefinementAssertion = false;
            TimedRefinementClockCeiling = PARUltility.TIME_REFINEMENT_CLOCK_CEILING;
            TimedRefinementClockFloor = PARUltility.TIME_REFINEMENT_CLOCK_FLOOR;
            Parameters = new HashSet<string>();
        }
    }

}
