using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class TimedInterrupt : PARProcess
    {
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;
        public Expression DelayTime;
        public byte TimerID;

        public TimedInterrupt(PARProcess firstProcess, PARProcess secondProcess, Expression delay, byte timerID)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            DelayTime = delay;
            TimerID = timerID;
            // TO DO: get closer to RTS
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.INTERRUPT + "[" + DelayTime.ExpressionID + "]" + SecondProcess.ProcessID);
        }



        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;

            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                vm.Process = new TimedInterrupt(vm.Process, SecondProcess, DelayTime, TimerID);
            }

            Constraint constraint = eStep.TheConstraint.Clone();
            constraint.AddInequality(TimerID, TimerOperationType.Equals, value);
            steps.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[interrupt[" + value + "]]", eStep.GlobalEnv, false, false, constraint));
            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            //get the int constant delay value
            //int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;
            List<PARConfigurationWithChannelData> steps = FirstProcess.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                //DBM newDBM = vm.DBM.Clone();
                //newDBM.AddConstraint(TimerID, TimerOperationType.LessThanOrEqualTo, value);
                //vm.DBM = newDBM;

                vm.Process = new TimedInterrupt(vm.Process, SecondProcess, DelayTime, TimerID);
                //toReturn.Add(vm);
            }
            return steps;
            //return toReturn;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            //get the int constant delay value
            //int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;

            //List<RTSConfiguration> toReturn = new List<RTSConfiguration>();

            List<PARConfiguration> steps = FirstProcess.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                //DBM newDBM = vm.DBM.Clone();
                //newDBM.AddConstraint(TimerID, TimerOperationType.LessThanOrEqualTo, value);
                //vm.DBM = newDBM;
                vm.Process = new TimedInterrupt(vm.Process, this.SecondProcess, DelayTime, TimerID);
                //toReturn.Add(vm);
            }

            return steps;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList<string>(Variables, SecondProcess.GetGlobalVariables());
            Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess.ToString() + " interrupt[" + DelayTime + "] " + SecondProcess.ToString() + ")";
            }
            else
            {
                return "(" + FirstProcess.ToString() + " interrupt[" + DelayTime + "]" + Constraint.ClockToString(TimerID) + " " + SecondProcess.ToString() + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new TimedInterrupt(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), DelayTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());//eStep));
        //    Common.Classes.Ultility.Ultility.Union(timers, SecondProcess.GetActiveTimer());//eStep));

        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddInequality(TimerID, DelayTime);
                    return new TimedInterrupt(process, SecondProcess, DelayTime, TimerID);
                }

                clocks.AddInequality(timerID, DelayTime);
                return new TimedInterrupt(process, SecondProcess, DelayTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, DelayTime);
                return null;
            }

            clocks.AddInequality(timerID, DelayTime);
            return new TimedInterrupt(FirstProcess, SecondProcess, DelayTime, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}