﻿using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class WaitProcess : PARProcess
    {
        //timer id is initially 0, once it is assigned, then it > 0;
        public byte TimerID;
        public Expression DelayTime;

        public WaitProcess(Expression delay, byte timerID)
        {
            DelayTime = delay;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(Constants.WAIT + "[" + delay.ExpressionID + "]");
        }




        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            List<PARConfiguration> toReturn = new List<PARConfiguration>(1);

            //get the int constant delay value
            int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;
            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value);

            Constraint newConstraint = eStep.TheConstraint.Clone();
            newConstraint.AddInequality(TimerID, TimerOperationType.Equals, value);
            PARConfiguration step = new PARConfiguration(new Skip(), Constants.TAU, "[Wait[" + value + "]]", eStep.GlobalEnv, false, false, newConstraint);
            toReturn.Add(step);
            return toReturn;
        }




        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "Wait[" + DelayTime + "]";
            }
            else
            {
                return "Wait[" + DelayTime + "]" + Constraint.ClockToString(TimerID);
            }
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new WaitProcess(DelayTime.ClearConstant(constMapping), TimerID);
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>(1);
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }
            
        //    return timers;
        //}

        public override List<string> GetGlobalVariables()
        {
            return DelayTime.GetVars();
        }

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, DelayTime);
                return null;
            }
            clocks.AddInequality(timerID, DelayTime);
            return new WaitProcess(DelayTime, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}
