using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class WithinProcess : PARProcess
    {
        public byte TimerID;
        public  PARProcess FirstProcess;
        public  Expression DelayTime;

        public WithinProcess(PARProcess firstProcess, Expression delay, byte timerID)
        {
            FirstProcess = firstProcess;
            DelayTime = delay;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.WITHIN + "[" + DelayTime.ExpressionID + "]");
        }




        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;
            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new WithinProcess(vm.Process, DelayTime, TimerID);
                }
            }

            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            return FirstProcess.SyncOutput(eStep);
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            return FirstProcess.SyncInput(eStep);
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            return FirstProcess.GetChannels();
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess.ToString() + " within[" + DelayTime + "])";
            }
            else
            {
                return "(" + FirstProcess.ToString() + " within[" + DelayTime + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return FirstProcess.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new WithinProcess(FirstProcess.ClearConstant(constMapping), DelayTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);    
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());
            
        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddInequality(TimerID, DelayTime);
                    return new WithinProcess(process, DelayTime, TimerID);
                }
                clocks.AddInequality(timerID, DelayTime);
                return new WithinProcess(process, DelayTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, DelayTime);
                return null;
            }

            clocks.AddInequality(timerID, DelayTime);
            return new WithinProcess(FirstProcess, DelayTime, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return FirstProcess.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}