using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class WithinRangeProcess : PARProcess
    {
        public byte TimerID;
        public PARProcess FirstProcess;
        public Expression FromTime;
        public Expression ToTime;

        public WithinRangeProcess(PARProcess firstProcess, Expression from, Expression to, byte timerID)
        {
            FirstProcess = firstProcess;
            FromTime = from;
            ToTime = to;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.WITHINRANGE + "[" + FromTime.ExpressionID + "," + ToTime.ExpressionID + "]");
        }




        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;
            int value2 = (EvaluatorDenotational.Evaluate(ToTime, eStep.GlobalEnv) as IntConstant).Value;

            eStep.AddInequality(TimerID, TimerOperationType.LessThanOrEqualTo, value2);

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                if (vm.Event == Constants.TAU)
                {
                    vm.Process = new WithinRangeProcess(vm.Process, FromTime, ToTime, TimerID);
                }
                else
                {
                    // ETIENNE SAYS: REDUNDANT CODE!
                    Constraint newConstraint = vm.TheConstraint.Clone();
                    newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
                    vm.TheConstraint = newConstraint;
                }
            }

            return steps;
        }




        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;
            List<PARConfigurationWithChannelData> steps = FirstProcess.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                // ETIENNE SAYS: REDUNDANT CODE!
                Constraint newConstraint = vm.TheConstraint.Clone();
                newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
                vm.TheConstraint = newConstraint;
            }

            return steps; 
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            //get the int constant delay value
            int value1 = (EvaluatorDenotational.Evaluate(FromTime, eStep.GlobalEnv) as IntConstant).Value;

            List<PARConfiguration> steps = FirstProcess.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                // ETIENNE SAYS: REDUNDANT CODE!
                Constraint newConstraint = vm.TheConstraint.Clone();
                newConstraint.AddInequality(TimerID, TimerOperationType.GreaterThanOrEqualTo, value1);
                vm.TheConstraint = newConstraint;
            }

            return steps;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList(Variables, FromTime.GetVars());
            Common.Classes.Ultility.Ultility.AddList(Variables, ToTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            return FirstProcess.GetChannels();
        }

        public override string ToString()
        {
            if (TimerID == 0)
            {
                return "(" + FirstProcess + " within[" + FromTime + "," + ToTime + "])";
            }
            else
            {
                return "(" + FirstProcess + " within[" + FromTime + "," + ToTime + "]" + Constraint.ClockToString(TimerID) + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            return FirstProcess.GetAlphabets(visitedDefinitionRefs);
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new WithinRangeProcess(FirstProcess.ClearConstant(constMapping), FromTime.ClearConstant(constMapping), ToTime.ClearConstant(constMapping), TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);    
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());
            
        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddInequality(TimerID, ToTime);
                    return new WithinRangeProcess(process, FromTime, ToTime, TimerID);
                }

                clocks.AddInequality(timerID, ToTime);
                return new WithinRangeProcess(process, FromTime, ToTime, timerID);
            }

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddInequality(TimerID, ToTime);
                return null;
            }

            clocks.AddInequality(timerID, ToTime);
            return new WithinRangeProcess(FirstProcess, FromTime, ToTime, timerID);
        }

        public override PARProcess GetTopLevelConcurrency(List<string> visitedDef)
        {
            return FirstProcess.GetTopLevelConcurrency(visitedDef);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}