using System.Collections.Generic;
using System.Text;
using PAT.Common.Classes.Assertion;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.LTS;
using System.Text.RegularExpressions;
using PAT.Common.Classes.ModuleInterface;
using PAT.PAR.LTS;
using Sequence = PAT.PAR.LTS.Sequence;

namespace PAT.PAR.Ultility
{
    public sealed class PARUltility
    {
        public static bool AllowGuardInPAR = false;
        public static bool TAU_URGENT = false;
        public static int TIME_REFINEMENT_CLOCK_CEILING = 32767;
        public static int TIME_REFINEMENT_CLOCK_FLOOR = -32768;
        public static bool LINK_PAR = false;


        //public static bool Visit(StringDictionary<List<Constraint>> Visited, PARConfiguration configuration)
        //{
        //    string state = configuration.GetIDNoConstraint();
        //    Constraint constraint = configuration.TheConstraint;
        //    if (Visited.ContainsKey(state))
        //    {
        //        List<Constraint> constraints = Visited.GetContainsKey(state);
        //        int size = constraints.Count;
        //        int i = 0;
        //        while (i < constraints.Count)
        //        {
        //            Constraint constrainti = constraints[i];
        //            if (constraint.IsSubSetOf(constrainti))
        //            {
        //                return true;
        //            }

        //            if (constrainti.IsSubSetOf(constraint))
        //            {
        //                constraints.RemoveAt(i);
        //            }
        //            else
        //            {
        //                i++;
        //            }
        //        }

        //        constraints.Add(constraint);
        //        Visited.ExtraCounter += constraints.Count - size;
        //        return false;
        //    }
        //    else
        //    {
        //        List<Constraint> constraints = new List<Constraint>();
        //        constraints.Add(constraint);
        //        Visited.Add(state, constraints);
        //        return false;
        //    }
        //}
        


        public static string GenerateLaTexString(Specification specification)
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendLine(@"\usepackage{pat}");

            if (specification.SpecValuation.Variables != null)
            {
                sb.AppendLine();
                sb.AppendLine("//====================Global Variable Definitions====================");

                sb.AppendLine(@"\begin{syntax}");

                foreach (StringDictionaryEntryWithKey<ExpressionValue> pair in specification.SpecValuation.Variables._entries)
                {
                    if (pair != null)
                    {
                        string range = "";
                        if (Valuation.VariableLowerBound.ContainsKey(pair.Key))
                        {
                            range = ":\\{" + Valuation.VariableLowerBound.GetContainsKey(pair.Key) + "..";
                        }

                        if (Valuation.VariableUpperLowerBound.ContainsKey(pair.Key))
                        {
                            if (range == "")
                            {
                                range = ":\\{.." + Valuation.VariableUpperLowerBound.GetContainsKey(pair.Key) + "\\}";
                            }
                            else
                            {
                                range += Valuation.VariableUpperLowerBound.GetContainsKey(pair.Key) + "\\}";
                            }
                        }
                        else
                        {
                            if (range != "")
                            {
                                range += "\\}";
                            }
                        }

                        sb.AppendLine("var~" + pair.Key + range + " & = &" + pair.Value + @";\\");
                    }
                }

                sb.AppendLine(@"\end{syntax}");
            }

            if (specification.ChannelDatabase.Count > 0)
            {
                sb.AppendLine();
                sb.AppendLine("//====================Channel Definitions====================");
                sb.AppendLine(@"\begin{syntax}");
                foreach (KeyValuePair<string, ChannelQueue> pair in specification.ChannelDatabase)
                {
                    sb.AppendLine("channel ~ " + pair.Key + " ~ " + pair.Value.Size + @";\\");
                }
                sb.AppendLine(@"\end{syntax}");
            }

            sb.AppendLine("//====================Process Definitions====================");
            sb.AppendLine(@"\begin{syntax}");
            foreach (Definition entry in specification.DefinitionDatabase.Values)
            {
                sb.AppendLine(entry.Name + "(" + Common.Classes.Ultility.Ultility.PPStringList(entry.LocalVariables) + ") & = & " + GenerateProcessLaTexString(entry.Process) + @";\\");
            }
            sb.AppendLine(@"\end{syntax}");

            //if (Specification.AlphaDatabase.Count > 0)
            //{
            //    sb.AppendLine("\r\n//============================Alphabets Definitions============================");
            //    sb.AppendLine(@"\begin{syntax}");
            //    foreach (KeyValuePair<string, EventCollection> pair in Specification.AlphaDatabase)
            //    {
            //        sb.AppendLine("\\#alphabet ~ " + pair.Key + " \\{" + Common.Classes.Ultility.Ultility.PPStringList(pair.Value) + "\\};");
            //    }
            //    sb.AppendLine(@"\end{syntax}");
            //}

            if (specification.DeclarationDatabase.Count > 0)
            {
                sb.AppendLine("\r\n//============================Declarations============================");
                sb.AppendLine(@"\begin{syntax}");
                foreach (KeyValuePair<string, Expression> entry in specification.DeclarationDatabase)
                {
                    sb.AppendLine("\\#define ~ " + entry.Key + " ~ " + Common.Ultility.Ultility.GenerateLaTexString(entry.Value) + @";\\");
                }
                sb.AppendLine(@"\end{syntax}");
            }

            if (specification.AssertionDatabase.Count > 0)
            {
                sb.AppendLine("\r\n//============================Asserstion Definitions============================");
                sb.AppendLine(@"\begin{syntax}");
                foreach (KeyValuePair<string, AssertionBase> entry in specification.AssertionDatabase)
                {
                    string replacedKey = Regex.Replace(entry.Key, "refines\\s*?<FD>", "\\sqsupseteq_{FD}");
                    replacedKey = Regex.Replace(replacedKey, "refines\\s*?<F>", "\\sqsupseteq_F");
                    sb.AppendLine("\\#assert ~ " + replacedKey.Replace("[]", " \\always ").Replace("<>", " \\eventually ").Replace("->", " \\fun ").Replace("|=", " \\models ").Replace("deadlockfree", " ~deadlockfree").Replace("reaches", " ~reaches~").Replace("refines", "\\sqsupseteq_T") + @";\\");

                }


                sb.AppendLine(@"\end{syntax}");
            }

            return sb.ToString();
        }

        public static string GenerateProcessLaTexString(PARProcess process)
        {
            string returnString = "";

            if (process is AtomicProcess)
            {
                AtomicProcess atomicProcess = process as AtomicProcess;
                return "atomic" + "\\{" + GenerateProcessLaTexString(atomicProcess.Process) + "\\}";
            }
            else if (process is CaseProcess)
            {
                CaseProcess caseP = process as CaseProcess;
                if (caseP.Processes.Length == 1)
                {
                    return "if " + Common.Ultility.Ultility.GenerateLaTexString(caseP.Conditions[0])
                        + " \\{" + GenerateProcessLaTexString(caseP.Processes[0]) + "\\}";
                }
                else
                {
                    StringBuilder s = new StringBuilder();
                    s.AppendLine("case \\{");
                    for (int i = 0; i < caseP.Processes.Length; i++)
                    {
                        Expression con = caseP.Conditions[i];
                        PARProcess p = caseP.Processes[i];
                        s.AppendLine(Common.Ultility.Ultility.GenerateLaTexString(con) + ":" + GenerateProcessLaTexString(p));
                    }
                    s.AppendLine("\\}");
                    return s.ToString();
                }
            }
            else if (process is ChannelInput)
            {
                ChannelInput input = process as ChannelInput;

                string s = "";
                foreach (Expression item in input.ExpressionList)
                {
                    s += Common.Ultility.Ultility.GenerateLaTexString(item) + ",";
                }
                s = s.TrimEnd(',');

                return input.ChannelName + "?(" + s + ") \\fun " + GenerateProcessLaTexString(input.Process);
            }
            else if (process is ChannelInputGuarded)
            {
                ChannelInputGuarded inputGuarded = process as ChannelInputGuarded;

                string s = "";
                foreach (Expression item in inputGuarded.ExpressionList)
                {
                    s += Common.Ultility.Ultility.GenerateLaTexString(item) + ",";
                }
                s = s.TrimEnd(',');

                return inputGuarded.ChannelName + "?["
                       + Common.Ultility.Ultility.GenerateLaTexString(inputGuarded.GuardExpression) + "]"
                       + "(" + s + ") \\fun "
                       + GenerateProcessLaTexString(inputGuarded.Process);
            }
            else if (process is ChannelOutput)
            {
                ChannelOutput output = process as ChannelOutput;

                string s = "";
                foreach (Expression item in output.ExpressionList)
                {
                    s += Common.Ultility.Ultility.GenerateLaTexString(item) + ",";
                }
                s = s.TrimEnd(',');

                return output.ChannelName + "!(" + s + ") \\fun " + GenerateProcessLaTexString(output.Process);

            }
            else if (process is ConditionalChoice)
            {
                ConditionalChoice conditionalChoice = process as ConditionalChoice;
                return "if " + Common.Ultility.Ultility.GenerateLaTexString(conditionalChoice.ConditionalExpression) + " \\{" + GenerateProcessLaTexString(conditionalChoice.FirstProcess) + " \\} else \\{" + GenerateProcessLaTexString(conditionalChoice.SecondProcess) + "\\}";
            }
            //else if (process is ConditionalChoiceAtomic)
            //{
            //    ConditionalChoiceAtomic conditionalChoiceAtomic = process as ConditionalChoiceAtomic;
            //    return "ifa" + Common.Ultility.Ultility.GenerateLaTexString(conditionalChoiceAtomic.ConditionalExpression)
            //            + " \\{"
            //            + GenerateProcessLaTexString(conditionalChoiceAtomic.FirstProcess)
            //            + " \\} else \\{"
            //            + GenerateProcessLaTexString(conditionalChoiceAtomic.SecondProcess)
            //            + "\\}";
            //}
            else if (process is DataOperationPrefix)
            {
                DataOperationPrefix prefix = process as DataOperationPrefix;
                returnString += GenerateProcessLaTexString(prefix.Event) + "\\{" + prefix.AssignmentExpr + "\\} \\fun " + GenerateProcessLaTexString(prefix.Process);
            }
            else if (process is DefinitionRef)
            {
                DefinitionRef def = process as DefinitionRef;
                return def.Name + "(" + Common.Classes.Ultility.Ultility.PPStringList(def.Args) + ")";
            }
            else if (process is EventPrefix)
            {
                EventPrefix prefix = process as EventPrefix;

                returnString += GenerateProcessLaTexString(prefix.Event) + " \\fun " + GenerateProcessLaTexString(prefix.Process);
            }
            //else if (process is GuardProcess)
            //{
            //    GuardProcess guard = process as GuardProcess;
            //    returnString = "[" + Common.Ultility.Ultility.GenerateLaTexString(guard.Condition) + "]" + GenerateProcessLaTexString(guard.Process);
            //}
            else if (process is Hiding)
            {
                Hiding hiding = process as Hiding;
                return "(" + GenerateProcessLaTexString(hiding.Process) + @"\hide \{" + LaTexPPStringListDot(hiding.HidingAlphabets) + @"\})";
            }
            else if (process is IndexChoice)
            {
                IndexChoice choice = process as IndexChoice;

                if (choice.Processes == null)
                {
                    return "(\\extchoice " + choice.IndexedProcessDefinition.ToString() + ")";
                }

                string result = "(" + GenerateProcessLaTexString(choice.Processes[0]);
                for (int i = 1; i < choice.Processes.Count; i++)
                {
                    result += "\\extchoice ";
                    result += GenerateProcessLaTexString(choice.Processes[i]);
                }
                result += ")";
                return result;

            }
            else if (process is IndexExternalChoice)
            {
                IndexExternalChoice choice = process as IndexExternalChoice;

                if (choice.Processes == null)
                {
                    return "(\\extchoice " + choice.IndexedProcessDefinition.ToString() + ")";
                }

                string result = "(" + GenerateProcessLaTexString(choice.Processes[0]);
                for (int i = 1; i < choice.Processes.Count; i++)
                {
                    result += " [*] ";
                    result += GenerateProcessLaTexString(choice.Processes[i]);
                }
                result += ")";
                return result;
            }

            else if (process is IndexInterleave)
            {
                IndexInterleave interleave = process as IndexInterleave;

                if (interleave.Processes == null)
                {
                    return "(\\interleave " + interleave.IndexedProcessDefinition.ToString() + ")";
                }
                else
                {

                    StringBuilder s = new StringBuilder();

                    for (int i = 0; i < interleave.Processes.Count; i++)
                    {
                        PARProcess p = interleave.Processes[i];
                        s.Append(GenerateProcessLaTexString(p) + " \\interleave ");
                    }
                    return s.ToString().Substring(0, s.Length - 12);
                }
            }

            else if (process is IndexInterleaveAbstract)
            {
                IndexInterleaveAbstract interleaveAbstract = process as IndexInterleaveAbstract;

                string toReturn = "";

                if (interleaveAbstract.ProcessesCounter == null)
                {
                    foreach (PARProcess processe in interleaveAbstract.Processes)
                    {
                        if (interleaveAbstract.ProcessesActualSize[processe.ProcessID] == -1)
                        {
                            toReturn += @"(\interleave \{..\} @" + GenerateProcessLaTexString(processe) + ") \\interleave ";
                        }
                        else
                        {
                            toReturn += @"(\interleave \{" + interleaveAbstract.ProcessesActualSize[processe.ProcessID] + "\\} @" + GenerateProcessLaTexString(processe) + ") \\interleave ";
                        }
                    }
                }
                else
                {
                    foreach (PARProcess processe in interleaveAbstract.Processes)
                    {
                        if (interleaveAbstract.ProcessesCounter[processe.ProcessID] == -1)
                        {
                            toReturn += @"(\interleave \{..\} $@$" + GenerateProcessLaTexString(processe) + ") \\interleave ";
                        }
                        else
                        {
                            toReturn += @"(\interleave \{" + interleaveAbstract.ProcessesCounter[processe.ProcessID] + "\\} $@$" + GenerateProcessLaTexString(processe) + ") \\interleave ";
                        }
                    }
                }

                return toReturn.Substring(0, toReturn.Length - 12);
            }
            else if (process is IndexInternalChoice)
            {
                IndexInternalChoice parallel = process as IndexInternalChoice;

                if (parallel.Processes == null)
                {
                    returnString += "(\\intchoice " + parallel.IndexedProcessDefinition.ToString() + ")";
                }
                else
                {
                    for (int i = 0; i < parallel.Processes.Count; i++)
                    {
                        PARProcess p = parallel.Processes[i];
                        if (i != parallel.Processes.Count - 1)
                        {
                            returnString += GenerateProcessLaTexString(p) + " \\intchoice ";
                        }
                        else
                        {
                            returnString += GenerateProcessLaTexString(p);
                        }
                    }
                }
            }
            else if (process is IndexParallel)
            {
                IndexParallel parallel = process as IndexParallel;

                if (parallel.Processes == null)
                {
                    returnString += "(\\parallel " + parallel.IndexedProcessDefinition.ToString() + ")";

                }
                else
                {

                    for (int i = 0; i < parallel.Processes.Count; i++)
                    {
                        PARProcess p = parallel.Processes[i];
                        if (i != parallel.Processes.Count - 1)
                        {
                            returnString += GenerateProcessLaTexString(p) + " \\parallel ";
                        }
                        else
                        {
                            returnString += GenerateProcessLaTexString(p);
                        }
                    }
                }
            }
            else if (process is Interrupt)
            {
                Interrupt interrupt = process as Interrupt;
                return "(" + GenerateProcessLaTexString(interrupt.FirstProcess) + " \\interrupt " + GenerateProcessLaTexString(interrupt.SecondProcess) + ")";
            }
            else if (process is Sequence)
            {
                Sequence sequence = process as Sequence;
                return "(" + GenerateProcessLaTexString(sequence.FirstProcess) + " ; " + GenerateProcessLaTexString(sequence.SecondProcess) + ")";
            }
            else if (process is Skip)
            {
                return "Skip";
            }
            else if (process is Stop)
            {
                return "Stop";
            }
            else if (process is DeadlineProcess)
            {
                DeadlineProcess deadline = process as DeadlineProcess;

                return "(" + GenerateProcessLaTexString(deadline.Process) + " deadline["
                        + Common.Ultility.Ultility.GenerateLaTexString(deadline.DelayTime) + "]";

            }
            else if (process is TimedInterrupt)
            {
                TimedInterrupt timedInterrupt = process as TimedInterrupt;
                return "(" + GenerateProcessLaTexString(timedInterrupt.FirstProcess)
                    + " interrupt["
                    + Common.Ultility.Ultility.GenerateLaTexString(timedInterrupt.DelayTime) + "]"
                    + GenerateProcessLaTexString(timedInterrupt.SecondProcess) + ")";
            }
            else if (process is TimeOutProcess)
            {
                TimeOutProcess timeOutProcess = process as TimeOutProcess;
                return "(" + GenerateProcessLaTexString(timeOutProcess.FirstProcess)
                    + " timeout[" + Common.Ultility.Ultility.GenerateLaTexString(timeOutProcess.DelayTime) + "]"
                    + GenerateProcessLaTexString(timeOutProcess.SecondProcess) + ")";
            }
            else if (process is WaitProcess)
            {
                WaitProcess waitProcess = process as WaitProcess;
                return "Wait[" + Common.Ultility.Ultility.GenerateLaTexString(waitProcess.DelayTime) + "]";
            }

            //else if (process is WaitRangeProcess)
            //{
            //    WaitRangeProcess waitRangeProcess = process as WaitRangeProcess;

            //    return "Wait[" + Common.Ultility.Ultility.GenerateLaTexString(waitRangeProcess.FromTime) + ","
            //                   + Common.Ultility.Ultility.GenerateLaTexString(waitRangeProcess.ToTime) + "]";
            //}
            else if (process is WaituntilProcess)
            {
                WaituntilProcess waitUtilProcess = process as WaituntilProcess;

                return "(" + GenerateProcessLaTexString(waitUtilProcess.Process) + " waituntil[" +
                        Common.Ultility.Ultility.GenerateLaTexString(waitUtilProcess.DelayTime) + "])";

            }
            else if (process is WithinProcess)
            {
                WithinProcess withinProcess = process as WithinProcess;

                return "(" + GenerateProcessLaTexString(withinProcess.FirstProcess) + " within["
                        + Common.Ultility.Ultility.GenerateLaTexString(withinProcess.DelayTime) + "])";
            }
            else if (process is WithinRangeProcess)
            {
                WithinRangeProcess withinRangeProcess = process as WithinRangeProcess;

                return "(" + GenerateProcessLaTexString(withinRangeProcess.FirstProcess) + " within["
                        + Common.Ultility.Ultility.GenerateLaTexString(withinRangeProcess.FromTime) + ","
                        + Common.Ultility.Ultility.GenerateLaTexString(withinRangeProcess.ToTime) + "])";

            }

            return returnString;
        }

        public static string GenerateProcessLaTexString(Event evt)
        {
            string toReturn;
            if (evt.EventID == null)
            {
                toReturn = evt.BaseName + LaTexPPStringListDot(evt.ExpressionList);
            }
            else
            {
                toReturn = evt.EventID;
            }

            return toReturn.Replace("\u03C4", "\\tau");

            //switch (evt.FairnessLabelType)
            //{
            //    case FairnessLabelType.Normal:
            //        return toReturn;
            //    case FairnessLabelType.StrongLive:
            //        return "sl(" + toReturn + ")";
            //    case FairnessLabelType.WeakLive:
            //        return "wl(" + toReturn + ")";
            //    case FairnessLabelType.StrongFair:
            //        return "sf(" + toReturn + ")";
            //    case FairnessLabelType.WeakFair:
            //        return "wf(" + toReturn + ")";
            //    default:
            //        throw new Exception();
            //}
        }

        public static string LaTexPPStringListDot(Expression[] list)
        {
            if (list == null)
            {
                return "";
            }

            string s = "";
            foreach (Expression item in list)
            {
                s += "." + Common.Ultility.Ultility.GenerateLaTexString(item);
            }
            return s;
        }

        public static string LaTexPPStringListDot(List<Event> list)
        {
            if (list == null)
            {
                return "";
            }

            string s = "";
            foreach (Event item in list)
            {
                s += "." + GenerateProcessLaTexString(item);
            }
            return s;
        }
    }
}